from ...web.flask import instance_proxy


class PGConfPlugin:
    def __init__(self, app):
        self.app = app

    def load(self):
        __import__(__name__ + ".routes")
        instance_proxy.generic_proxy("/pgconf/configuration")
        instance_proxy.generic_proxy("/pgconf/configuration", method="POST")
        instance_proxy.generic_proxy("/pgconf/configuration/categories")
        instance_proxy.generic_proxy("/pgconf/configuration/status")
        instance_proxy.generic_proxy("/pgconf/configuration/category/<path:category>")



# Setup and run Prometheus for temBoard.
#
# This is a standalone Prometheus instance, for testing purpose only.
# It is not intended to be used in production.
#
# Until 9.0 the command is not exposed.
# Run with python -m temboardui.cli.prometheus.
#
# Requires a running temboard instance aside to expose metrics.
#
# Dev Grafana (http://localhost:3000) is setup to use this Prometheus as
# `temBoard prometheus` with a dedicated dashboard.
#
import logging
import os
import signal
import subprocess
import sys
from datetime import datetime

import jinja2

import temboardui

from ..model import Session, orm
from ..toolkit.app import SubCommand
from ..toolkit.errors import UserError
from ..toolkit.services import Service
from .app import app

logger = logging.getLogger(__package__ + ".prometheus")


@app.command
class Prometheus(SubCommand):
    """Standalone Prometheus for temBoard.

    For testing purpose only.

    """

    def main(self, args):
        if not self.app.config.monitoring.prometheus:
            raise UserError("missing prometheus binary")

        prometheus = PrometheusService(app=self.app)
        with prometheus:
            prometheus.run()


class PrometheusService(Service):
    def __init__(self, **kwargs):
        super().__init__(name="prometheus manager", **kwargs)
        self.proc = None

    @property
    def pidfile(self):
        return self.app.config.temboard.home + "/prometheus.pid"

    @property
    def home(self):
        return self.app.config.temboard.home + "/prometheus"

    def serve(self):
        self.terminate_running_prometheus(self.pidfile)

        logger.info("Provisionning Prometheus in %s.", self.home)
        provision_prometheus(self.app, self.home)
        logger.info("Starting %s.", self.app.config.monitoring.prometheus)
        cmd = [
            self.app.config.monitoring.prometheus,
            f"--config.file={self.home}/prometheus.yml",
            # 0.0.0.0 is reachable from dev grafana for testing purpose.
            # Once we have alertmanager, use localhost.
            "--web.listen-address=0.0.0.0:8890",
            "--log.level=debug",
            "--storage.tsdb.retention.time=1h",
        ]
        self.proc = subprocess.Popen(cmd, cwd=self.home)
        logger.info("Started Prometheus as PID %s.", self.proc.pid)
        with open(self.pidfile, "w") as fo:
            fo.write(str(self.proc.pid))
        self.setup_instances()
        logger.debug("Waiting prometheus service forever.")
        self.proc.wait()

    def terminate_running_prometheus(self, pidfile):
        if not os.path.exists(pidfile):
            return

        with open(pidfile) as fo:
            pid = fo.read()

        try:
            logger.debug("Eventually terminating Prometheus PID %s.", pid)
            os.kill(int(pid), signal.SIGTERM)
            logger.info("Prometheus running as PID %s. Terminating.", pid)
        except OSError:
            logger.debug("Spurious PID file %s.", pidfile)

    def setup_instances(self):
        session = Session()
        for (instance,) in session.execute(orm.Instance.all()):
            if not instance.discover:
                logger.debug("Skipping unreachable instance %s.", instance)
                continue
            if instance.discover["temboard"]["agent_version"] < "8":
                logger.debug("Skipping old agent %s.", instance)
                continue
            provision_instance(self.app, instance)
            self.proc.send_signal(signal.SIGHUP)

    def reload(self):
        super().reload()
        if self.proc:
            self.proc.send_signal(signal.SIGHUP)

    def teardown(self):
        if not self.proc:
            return
        logger.info("Stopping Prometheus.")
        self.proc.terminate()
        self.proc = None
        os.unlink(self.app.config.temboard.home + "/prometheus.pid")


def provision_prometheus(app, home):
    logger.debug("Ensure prometheus home is created.")
    os.makedirs(home + "/instances.d", exist_ok=True)
    t = jinja2.Template(
        source=PROMETHEUS_CONFIG_TEMPLATE,
        trim_blocks=True,
        lstrip_blocks=True,
        undefined=jinja2.StrictUndefined,
        keep_trailing_newline=True,
    )
    conf = t.render(
        app=app,
        apikey="UNSECURE_DEV_APIKEY",
        home=home,
        now=datetime.utcnow(),
        temboard_version=temboardui.__version__,
    )
    with open(home + "/prometheus.yml", "w") as fo:
        os.chmod(fo.name, 0o600)
        fo.write(conf)


PROMETHEUS_CONFIG_TEMPLATE = """\
# This file is generated by temBoard. Do not edit.
#
# Generation Date: {{ now }}
# temBoard Version: {{ temboard_version }}
#
scrape_configs:
- job_name: temboard
  scrape_interval: 60s
  scheme: http
  authorization:
    type: Bearer
    credentials: {{ apikey }}
  file_sd_configs:
  - files:
    - instances.d/instance-*.yml
"""


def provision_instance(app, instance):
    logger.info("Provisionning %s.", instance)
    t = jinja2.Template(
        source=INSTANCE_CONFIG_TEMPLATE,
        trim_blocks=True,
        lstrip_blocks=True,
        undefined=jinja2.StrictUndefined,
        keep_trailing_newline=True,
    )
    conf = t.render(
        app=app,
        instance=instance,
        now=datetime.utcnow(),
        temboard_version=temboardui.__version__,
    )
    path = "{}/prometheus/instances.d/instance-{}-{}.scrape.yml".format(
        app.config.temboard.home, instance.agent_address, instance.agent_port
    )
    with open(path, "w") as fo:
        os.chmod(fo.name, 0o600)
        fo.write(conf)


INSTANCE_CONFIG_TEMPLATE = """\
# This file is generated by temBoard. Do not edit.
#
# Generation Date: {{ now }}
# temBoard Version: {{ temboard_version }}
#
- targets: [localhost:8888]
  labels:
    __metrics_path__: "/proxy/{{ instance.agent_address }}/{{ instance.agent_port }}/monitoring/metrics"
    instance:  "{{ instance.hostname }}:{{ instance.pg_port }}"
    agent:  "{{ instance.agent_address }}:{{ instance.agent_port }}"
"""  # noqa


if "__main__" == __name__:
    from ..__main__ import main

    sys.exit(main(argv=["prometheus"] + sys.argv[1:]))



import functools
import json
import logging
import os
from csv import writer as CSVWriter
from io import StringIO

from tornado import web as tornadoweb
from tornado.concurrent import run_on_executor
from tornado.escape import json_decode, json_encode, url_escape
from tornado.gen import coroutine
from tornado.template import Loader as TemplateLoader
from tornado.web import Application as TornadoApplication
from tornado.web import HTTPError, RequestHandler

from ..agentclient import TemboardAgentClient
from ..application import get_instance, get_role_by_cookie
from ..errors import TemboardUIError
from ..model import Session as DBSession
from ..toolkit.perf import PerfCounters
from ..toolkit.utils import JSONEncoder, utcnow

logger = logging.getLogger(__name__)


def admin_required(func):
    # Similar to flask_security.roles_required, but limited to admin role.
    func.__admin_required = True
    return func


def anonymous_allowed(func):
    # Reverse of flask_security.login_required.
    #
    # In temboard, very few pages are anonymous. Thus we have implicit
    # login_required. This behaviour can be disabled by using
    # @anonymous_allowed.
    func.__anonymous_allowed = True
    return func


def serialize_querystring(query):
    return "&".join(
        [
            f"{url_escape(name)}={url_escape(value)}"
            for name, value in sorted(query.items())
        ]
    )


class Response:
    def __init__(self, status_code=200, headers=None, secure_cookies=None, body=""):
        self.status_code = status_code
        self.headers = headers or {}
        self.secure_cookies = secure_cookies or {}
        self.body = body


class Redirect(Response, Exception):
    def __init__(self, location, permanent=False, secure_cookies=None):
        super().__init__(
            status_code=301 if permanent else 302,
            headers={"Location": location},
            body="Redirected to %s" % location,
            secure_cookies=secure_cookies,
        )


class TemplateRenderer:
    # Flask-like HTML render function, without thread local.

    GLOBAL_NAMESPACE = {}

    def __init__(self, path):
        self.loader = TemplateLoader(path)

    def __call__(self, template, **data):
        data = dict(self.GLOBAL_NAMESPACE, **data)
        # unsafe-eval is for jquery. unsafe-inline because we have
        # script tags in templates.
        csp = "default-src 'self' 'unsafe-inline' 'unsafe-eval' data:"
        if "VITEJS" in os.environ:
            csp += " localhost:5173 ws:"
        return Response(
            body=self.loader.load(template).generate(**data),
            headers={
                "Content-Type": "text/html; charset=UTF-8",
                "Content-Security-Policy": csp,
            },
        )


template_path = os.path.realpath(__file__ + "/../../templates")
render_template = TemplateRenderer(template_path)


def csvify(data, status_code=200):
    if isinstance(data, list):
        fo = StringIO()
        writer = CSVWriter(fo)
        for row in data:
            writer.writerow(row)
        data = fo.getvalue()
    elif not isinstance(data, (str, bytes)):
        raise ValueError("Malformed CSV data")
    return Response(
        status_code=status_code, headers={"Content-Type": "text/csv"}, body=data
    )


def jsonify(data, status_code=200):
    return Response(
        status_code=status_code,
        headers={"Content-Type": "application/json; charset=UTF-8"},
        body=json_encode(data),
    )


class CallableHandler(RequestHandler):
    # Adapt flask-like callable in Tornado Handler API.

    @property
    def executor(self):
        # To enable @run_on_executor methods, we must have executor property.
        return self.application.executor

    def compute_etag(self):
        # We don't want automatic caching for API.
        return None

    def initialize(self, callable_, blueprint=None, methods=None, logger=None):
        self.callable_ = callable_
        self.logger = logger or logging.getLogger(__name__)
        self.request.blueprint = blueprint
        self.request.config = self.application.config
        # run_on_executor searches for `executor` attribute of first argument.
        # Thus, we bind executor to request object.
        self.request.executor = self.executor
        self.request.handler = self
        self.SUPPORTED_METHODS = methods or ["GET"]

    def get_current_user(self):
        cookie = self.get_secure_cookie("temboard")
        if not cookie:
            return

        try:
            return get_role_by_cookie(self.db_session, cookie.decode("utf-8"))
        except Exception as e:
            self.logger.error("Failed to get role from cookie: %s ", e)

    @coroutine
    def get(self, *args, **kwargs):
        try:
            response = yield self.callable_(self.request, *args, **kwargs)
        except Redirect as r:
            response = r

        if response is None:
            response = ""
        if isinstance(response, (dict, str)):
            response = Response(body=response)
        self.write_response(response)

    # Let's use one handler for all supported methods.
    post = get
    delete = get

    def write_response(self, response):
        self.set_status(response.status_code)
        for k, v in list(response.headers.items()):
            if not isinstance(v, list):
                v = [v]
            self.clear_header(k)
            for v1 in v:
                self.add_header(k, v1)

        self.finish(response.body)


class Error404Handler(RequestHandler):
    def write_error(self, status_code, **kwargs):
        content_type = self.request.headers.get("Content-Type", "")
        if content_type.startswith("application/json"):
            self.write({"error": "Not found."})
        else:
            self.write("404: Not found.")


class DatabaseHelper:
    @classmethod
    def add_middleware(cls, func):
        @functools.wraps(func)
        def database_middleware(request, *args):
            request.db_session = request.handler.db_session = DBSession()
            try:
                response = func(request, *args)
            except Exception:
                # Expunge objects before rollback to implement
                # expire_on_rollback=False. This allow templates to reuse
                # request.instance object and joined object without triggering
                # lazy load.
                request.db_session.expunge_all()
                request.db_session.rollback()
                raise
            else:
                request.db_session.commit()
            finally:
                request.db_session.close()
                del request.db_session

            return response

        return database_middleware


class ErrorHelper:
    @classmethod
    def add_middleware(cls, func):
        @functools.wraps(func)
        def error_middleware(request, *args):
            try:
                return func(request, *args)
            except Redirect:
                raise
            except TemboardAgentClient.Error as e:
                code = e.response.status
                message = e.message
            except TemboardUIError as e:
                code = e.code
                message = e.message
            except HTTPError as e:
                code = e.status_code
                message = e.log_message
                if code == 404:
                    message += " Plugin may not be activated on agent side"
            except Exception as e:
                # Show traceback for developer, and HTML page for user.
                logger.exception("Unhandled Error:")
                code = 500
                message = str(e)

            logger.error("Request failed: %s %s.", code, message)
            return make_error(request, code, message)

        return error_middleware


class InstanceHelper:
    # This helper class implements all operations related to instance dedicated
    # request.

    INSTANCE_PARAMS = r"/(.+)/([0-9]{1,5})"
    PROXY_PREFIX = r"/proxy" + INSTANCE_PARAMS
    SERVER_PREFIX = r"/server" + INSTANCE_PARAMS

    @classmethod
    def add_middleware(cls, callable_):
        # Wraps an HTTP handler callable related to a Postgres instance

        @functools.wraps(callable_)
        def instance_middleware(request, address, port, *args):
            # Swallow adddress and port arguments.
            request.instance = cls(request)
            request.instance.fetch_instance(address, port)
            return callable_(request, *args)

        return instance_middleware

    def __init__(self, request):
        self.request = request

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.instance.hostname}>"

    def __str__(self):
        return f"{self.instance.hostname}:{self.instance.pg_port}"

    def check_active_plugin(self, name):
        """
        Ensure that the plugin is active for given instance
        """
        if name not in [p.plugin_name for p in self.instance.plugins]:
            raise HTTPError(408, "Plugin %s not activated." % name)

    def fetch_instance(self, address, port):
        self.instance = get_instance(self.request.db_session, address, port)
        self.instance.status = None
        if not self.instance:
            raise HTTPError(404)
        self.agent_id = "{}:{}".format(
            self.instance.agent_address, self.instance.agent_port
        )

    def fetch_status(self):
        try:
            data = self.request_agent("/status")
        except Exception as e:
            # agent is unreachable we forge a status response
            logger.error("Failed to fetch status: %s", e)
            data = {
                "temboard": {"status": "unreachable"},
                "postgres": {"status": "unreachable", "pending_restart": False},
                "system": {"status": "unreachable"},
            }
        self.instance.status = data

    def format_url(self, path=""):
        return f"/server/{self.agent_address}/{self.agent_port}{path}"

    def redirect(self, path):
        raise Redirect(location=self.format_url(path))

    def client(self):
        return TemboardAgentClient.factory(
            self.request.config,
            self.instance.agent_address,
            self.instance.agent_port,
            self.request.current_user.role_name,
        )

    def request_agent(self, path, method="GET", query=None, body=None):
        client = self.client()

        pathinfo = path
        if query:
            pathinfo += "?" + serialize_querystring(query)

        try:
            response = client.request(method=method, path=pathinfo, body=body)
            response.raise_for_status()
        except OSError as e:
            raise HTTPError(
                500,
                (
                    "Failed to contact agent %s:%s: %s. Is it running?"
                    % (self.instance.agent_address, self.instance.agent_port, e)
                ),
            )
        except ConnectionError as e:
            raise HTTPError(500, str(e))
        except TemboardAgentClient.Error as e:
            raise HTTPError(e.response.status, e.message)
        except Exception as e:
            logger.error("Proxied request failed: %s", e)
            raise HTTPError(500, "Unhandled error")
        else:
            return response.json()

    def get(self, *args, **kwargs):
        kwargs["method"] = "GET"
        return self.request_agent(*args, **kwargs)

    def post(self, *args, **kwargs):
        kwargs["method"] = "POST"
        return self.request_agent(*args, **kwargs)


def add_json_middleware(func):
    @functools.wraps(func)
    def json_middleware(request, *args):
        if "POST" == request.method:
            if request.body:
                try:
                    request.json = json_decode(request.body)
                except Exception as e:
                    raise HTTPError(400, str(e))
            else:
                logger.debug("Empty body for POST request.")
                request.json = None
        else:
            # Set empty body.
            request.json = None

        return func(request, *args)

    return json_middleware


def add_user_instance_middleware(func):
    # Ensures user is allowed to access to the instance
    @functools.wraps(func)
    def user_instance_middleware(request, *args):
        user = request.current_user

        if user is None:
            # Not logged in
            raise HTTPError(401, "Restricted area.")

        user_allowed = request.db_session.execute(
            request.instance.has_dba(user.role_name)
        ).scalar()
        if not user_allowed:
            raise HTTPError(403, "Restricted area.")

        return func(request, *args)

    return user_instance_middleware


class UserHelper:
    @classmethod
    def add_middleware(cls, func):
        @functools.wraps(func)
        def user_middleware(request, *args):
            role = request.current_user = request.handler.current_user

            anonymous_allowed = getattr(func, "__anonymous_allowed", False)
            if not anonymous_allowed and role is None:
                logger.debug("Redirecting anonymous to /login.")
                raise Redirect("/login")

            admin_required = getattr(func, "__admin_required", False)
            if admin_required and not role.is_admin:
                logger.debug("Refusing access to non-admin user.")
                raise HTTPError(403)

            return func(request, *args)

        return user_middleware


# Ensure @functools.wraps preserves User middleware attributes.
functools.WRAPPER_UPDATES += ("__admin_required", "__anonymous_allowed")


class Blueprint:
    def __init__(self, plugin_name=None):
        self.plugin_name = plugin_name
        self.rules = []
        self.perf = PerfCounters.setup(service="web", plugin=plugin_name)

    def add_rules(self, rules):
        self.rules.extend(rules)

    def generic_proxy(self, url, methods=None):
        # Pass-through implementation for /proxy/address/port/…
        url = r"(%s)" % url

        @self.instance_proxy(url, methods)
        def generic_instance_proxy(request, path):
            if request.blueprint and request.blueprint.plugin_name:
                request.instance.check_active_plugin(request.blueprint.plugin_name)
            path = url_escape(path, plus=False)
            # append raw query to path instead of rebuilding it in request_agent
            if request.query:
                path += "?" + request.query
            body = request.instance.request_agent(
                path=path, method=request.method, body=request.json
            )
            return jsonify(body)

    def instance_proxy(self, url, methods=None):
        # decorator for /proxy/address/port/… handlers.
        url = InstanceHelper.PROXY_PREFIX + url
        return self.route(url, methods=methods, with_instance=True, json=True)

    def instance_route(self, url, methods=None, **kwargs):
        # Helper to declare a route with instance URL prefix and middleware.
        return self.route(
            url=InstanceHelper.SERVER_PREFIX + url,
            methods=methods,
            with_instance=True,
            **kwargs,
        )

    def route(self, url, methods=None, with_instance=False, json=None):
        # Implements flask-like route registration of a simple synchronous
        # callable.

        # Enable JSON middleware on /json/ handlers.
        if json is None:
            json = url.startswith("/json/") or url.endswith(".json")

        def decorator(func):
            logger_name = func.__module__ + "." + func.__name__

            if with_instance:
                if url.startswith("/server/") or url.startswith("/proxy/"):
                    # Limit user/instance access control to /server/ and
                    # /proxy/.
                    # Admin area access control has already been performed
                    func = add_user_instance_middleware(func)

                func = InstanceHelper.add_middleware(func)
            func = UserHelper.add_middleware(func)

            if json:
                func = add_json_middleware(func)
            func = DatabaseHelper.add_middleware(func)
            func = ErrorHelper.add_middleware(func)

            @run_on_executor
            @functools.wraps(func)
            def sync_request_wrapper(request, *args):
                start = utcnow() if self.perf else None
                status = 500
                try:
                    response = func(request, *args)
                    if hasattr(response, "status_code"):
                        status = response.status_code
                    else:  # JSON data
                        status = 200
                    return response
                except (Redirect, HTTPError) as e:
                    status = e.status_code
                    raise
                except Exception as e:
                    # Since async traceback is useless, spit here traceback and
                    # forge simple HTTPError(500), no HTML error.
                    logger.exception("Unhandled Error:")
                    raise HTTPError(500, str(e))
                finally:
                    if self.perf:
                        response_time = utcnow() - start
                        instance_helper = getattr(request, "instance", None)
                        if with_instance and hasattr(instance_helper, "instance"):
                            agent = request.instance.agent_id
                        else:
                            agent = "undefined"

                        logger.debug(
                            "method=%s url=%s status=%s handler=%s"
                            " response_time=%s plugin=%s agent=%s service=web",
                            request.method,
                            request.path,
                            status,
                            logger_name,
                            response_time.total_seconds(),
                            self.plugin_name or "main",
                            agent,
                        )

            rules = [
                (
                    url,
                    CallableHandler,
                    dict(
                        blueprint=self,
                        callable_=sync_request_wrapper,
                        methods=methods or ["GET"],
                        logger=logging.getLogger(logger_name),
                    ),
                )
            ]
            self.add_rules(rules)
            return func

        return decorator


class WebApplication(TornadoApplication, Blueprint):
    def __init__(self, *a, **kwargs):
        super().__init__(*a, **kwargs)
        Blueprint.__init__(self)

    def configure(self, **settings):
        # Runtime configuration of application.
        #
        # This way, we can initialize app at import time to register handlers.
        # Then configure it at run time once configuration is parsed.

        self.settings.update(settings)

        # This comes from Tornado's __init__
        if self.settings.get("debug"):
            self.settings.setdefault("autoreload", True)
            self.settings.setdefault("compiled_template_cache", False)
            self.settings.setdefault("static_hash_cache", False)
            self.settings.setdefault("serve_traceback", True)

        self.start_time = utcnow()

    def add_rules(self, rules):
        if hasattr(self, "wildcard_router"):  # Tornado 4.5+
            self.wildcard_router.add_rules(rules)
        elif not self.handlers:
            self.add_handlers(r".*$", rules)
        else:
            rules = [tornadoweb.URLSpec(*r) for r in rules]
            self.handlers[0][1].extend(rules)


def make_error(request, code, message):
    # If ?noerror=1 is set, only return HTTP error code.
    if "1" == request.handler.get_argument("noerror", None):
        logger.debug("Hide error %s %s for ?noerror=1.", code, message)
        return Response(200)

    data = dict(code=code, error=message)

    if hasattr(request, "json"):
        return Response(code, body=data)

    if hasattr(request, "instance"):
        data["instance"] = request.instance

    template = "unauthorized.html" if 403 == code else "error.html"
    response = render_template(
        template, role=getattr(request, "current_user", "anonymous"), **data
    )
    response.status_code = code
    return response


# Change default cls argument to custom encoder.
if json.dumps.__kwdefaults__ is None:
    json.dumps.__kwdefaults__ = dict()
json.dumps.__kwdefaults__["cls"] = JSONEncoder


app = WebApplication()
# Hijack tornado.web access_log to log request in temboardui namespace.
tornadoweb.access_log = logging.getLogger(__package__)



#
# Somes notes on monitoring in temBoard
#
# Metrics are stored in different tables:
#
# - metric_*_current stores metrics one row per metric point
# - metric_*_30m_current is a compacted COPY of _current by interval
# - metric_*_history aggregates points per time interavl
#
# Tasks:
#
# - schedule_collector(): schedule a collector task for each agent in
#   inventory.
# - collector(host, port, key) inserts metrics history in metric_*_current
#   table.
# - history_tables_worker() move data from metric_*_current to
#   metric_*_history, grouped by time range. metric table is truncated
# - aggregate_data_worker() aggregates data in metric_*_30m_current and
#   metric_*_6h_current.
#

import logging
import os
import shutil
from datetime import datetime, timedelta

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest
from textwrap import dedent

import tornado.escape
import tornado.web
from psycopg2.extensions import AsIs
from sqlalchemy.exc import DataError, IntegrityError, ProgrammingError
from sqlalchemy.sql import text

from temboardui.agentclient import TemboardAgentClient
from temboardui.application import get_instance, send_mail, send_sms
from temboardui.model import worker_engine

from ...core import refresh_discover
from ...model import QUERIES, Session
from ...model import orm as coreorm
from ...toolkit import taskmanager
from ...toolkit import validators as v
from ...toolkit.configuration import OptionSpec
from ...toolkit.errors import UserError
from .alerting import check_specs
from .handlers import blueprint
from .model.db import insert_availability
from .model.orm import Check, CollectorStatus, Host, Instance
from .tools import (
    Stopwatch,
    check_preprocessed_data,
    get_host_id,
    get_instance_checks,
    get_instance_id,
    insert_metrics,
    merge_agent_info,
    populate_host_checks,
    preprocess_data,
    update_collector_status,
)

logger = logging.getLogger(__name__)
workers = taskmanager.WorkerSet()


class MonitoringPlugin:
    s = "monitoring"
    prometheus = shutil.which("prometheus")
    options_specs = [
        OptionSpec(s, "collect_max_duration", default=30, validator=int),
        OptionSpec(s, "prometheus", default=prometheus, validator=v.file_),
    ]

    def __init__(self, app):
        self.app = app
        self.app.config.add_specs(self.options_specs)

    def load(self):
        plugin_path = os.path.dirname(os.path.realpath(__file__))
        # Import Tornado handlers
        __import__(__name__ + ".handlers.alerting")
        __import__(__name__ + ".handlers.monitoring")
        # Import Flask routes
        __import__(__name__ + ".routes")
        self.app.tornado_app.add_rules(blueprint.rules)
        self.app.tornado_app.add_rules(
            [
                (
                    r"/js/monitoring/(.*)",
                    tornado.web.StaticFileHandler,
                    {"path": plugin_path + "/static/js"},
                )
            ]
        )
        self.app.worker_pool.add(workers)
        self.app.scheduler.add(workers)


@workers.schedule(id="aggregate_data", redo_interval=30 * 60)
@workers.register(pool_size=1)
def aggregate_data_worker(app):
    # Worker in charge of aggregate data
    stopwatch = Stopwatch()
    engine = worker_engine(app.config.repository)
    logger.info("Aggregating data.")
    with engine.connect() as conn:
        conn.execute("SET search_path TO monitoring")
        res = conn.execute("SELECT * FROM metric_tables_config()")
        (tables_config,) = res.fetchone()
        for config in tables_config.values():
            logger.debug("Aggregating data for metric %s.", config["name"])
            try:
                with conn.begin(), stopwatch:
                    res = conn.execute(
                        "SELECT * FROM aggregate_data_single(%s, %s, %s)",
                        (config["name"], config["record_type"], config["aggregate"]),
                    )
                    # Call here pg_sleep() using conn.execute() to fake slow
                    # aggregation.
                    table_name, nb_rows = res.fetchone()
                logger.debug(
                    "table=%s insert=%s timedelta=%s",
                    table_name,
                    nb_rows,
                    stopwatch.last_delta,
                )
            except Exception as e:
                logger.error("Failed to archive data: %s.", e)
                # search_path is lost on exception. Define it again.
                conn.execute("SET search_path TO monitoring")

    logger.debug("Monitoring data aggregation done.")
    logger.debug("Total time in SQL %s.", stopwatch.delta)


@workers.schedule(id="history_tables", redo_interval=3 * 60 * 60)  # 3h
@workers.register(pool_size=1)
def history_tables_worker(app):
    # Archive monitoring metric tables.
    #
    # Copy contents of every metric_*_current tables into corresponding
    # metric_*_history, aggregated. Then truncate _current tables.
    #
    # This task is triggered every 3 hours by monitoring_boostrap() below.
    #
    stopwatch = Stopwatch()
    engine = worker_engine(app.config.repository)
    logger.info("Archiving metrics.")
    with engine.connect() as conn:
        conn.execute("SET search_path TO monitoring")
        res = conn.execute("SELECT * FROM metric_tables_config()")
        (tables_config,) = res.fetchone()
        for config in tables_config.values():
            logger.debug("Archiving data for metric %s.", config["name"])
            try:
                with conn.begin(), stopwatch:
                    res = conn.execute(
                        "SELECT * FROM archive_current_metrics(%s, %s, %s)",
                        (config["name"], config["record_type"], config["history"]),
                    )
                    # Call here pg_sleep() using conn.execute() to fake slow
                    # archiving.
                    table_name, nb_rows = res.fetchone()
                logger.debug(
                    "table=%s insert=%s timedelta=%s",
                    table_name,
                    nb_rows,
                    stopwatch.last_delta,
                )
            except Exception as e:
                logger.error("Failed to archive data: %s.", e)
                # search_path is lost on exception. Define it again.
                conn.execute("SET search_path TO monitoring")

    logger.debug("Monitoring data archiving done.")
    logger.debug("Total time in SQL %s.", stopwatch.delta)


@workers.register(pool_size=10)
def check_data_worker(app, host_id, instance_id, data):
    # Worker in charge of checking preprocessed monitoring values
    worker_session = Session(bind=worker_engine(app.config.repository))

    check_preprocessed_data(app, worker_session, host_id, instance_id)

    worker_session.commit()
    worker_session.close()


@workers.schedule(id="purge_data", redo_interval=24 * 60 * 60)  # 24h
@workers.register(pool_size=1)
def purge_data_worker(app):
    """Background worker in charge of purging monitoring data. Purge policy
    is based on purge_after parameter from monitoring section. purger_after
    defines the number of day of data to keep, from now. Default value means
    there is no purge policy.
    """

    engine = worker_engine(app.config.repository)
    with engine.begin() as conn:
        logger.debug("Purging instances.")
        result = conn.execute(QUERIES["monitoring-purge-instances"])
        for deleted in result:
            logger.debug(
                "Instance purged. instance=%s:%s", deleted.hostname, deleted.port
            )
        if result.rowcount:
            logger.info("Purged deleted instances. count=%s", result.rowcount)

        logger.debug("Purging hosts.")
        result = conn.execute(QUERIES["monitoring-purge-hosts"])
        for deleted in result:
            logger.debug("Host purged. host=%ss", deleted.hostname)
        if result.rowcount:
            logger.info("Purged orphaned hosts. count=%s", result.rowcount)

    if not app.config.monitoring.purge_after:
        logger.info("No purge policy, end.")
        return

    logger.debug("Purging old data.")

    with engine.connect() as conn:
        # Get tablename list to purge from metric_tables_config()
        res = conn.execute(
            dedent("""
                SELECT
                    tablename
                FROM (
                    SELECT
                        tablename_prefix||'_'||suffix AS tablename
                    FROM
                        json_object_keys(monitoring.metric_tables_config()) AS tablename_prefix,
                        UNNEST(ARRAY['30m_current', '6h_current', 'current', 'history']) AS suffix
                ) AS q
                WHERE EXISTS (
                    SELECT 1
                    FROM
                        pg_catalog.pg_tables AS pgt
                    WHERE
                        pgt.tablename=q.tablename
                        AND pgt.schemaname = 'monitoring'
                )
                ORDER BY tablename;
            """)  # noqa
        )
        tablenames = [r["tablename"] for r in res.fetchall()]
        tablenames.extend(["state_changes", "check_changes"])

        purge_query_base = "DELETE FROM :tablename WHERE "

        for tablename in tablenames:
            # With history tables, we have to deal with tstzrange
            if tablename.endswith("_history"):
                query = (
                    purge_query_base
                    + "NOT (history_range && tstzrange(NOW() "
                    + "- ':nday days'::INTERVAL, NOW()))"
                )
            else:
                query = purge_query_base + "datetime < (NOW() - ':nday days'::INTERVAL)"

            logger.debug("Purging table %s", tablename)
            t = conn.begin()
            try:
                res_delete = conn.execute(
                    text(query),
                    tablename=AsIs("monitoring.%s" % tablename),
                    nday=app.config.monitoring.purge_after,
                )
                t.commit()
            except (ProgrammingError, IntegrityError) as e:
                logger.exception(e)
                logger.error("Could not delete data from table %s", tablename)
                t.rollback()
                continue

            if res_delete.rowcount > 0:
                logger.info("%s purged. count=%s", tablename, res_delete.rowcount)

    logger.debug("End of monitoring data purge worker.")


@workers.register(pool_size=1)
def notify_state_change(app, check_id, key, value, state, prev_state):
    # check if at least one notifications transport is configured
    # if it's not the case pass
    notifications_conf = app.config.notifications
    smtp_host = notifications_conf.smtp_host
    smtp_port = notifications_conf.smtp_port
    smtp_tls = notifications_conf.smtp_tls
    smtp_login = notifications_conf.smtp_login
    smtp_password = notifications_conf.smtp_password
    smtp_from_addr = notifications_conf.smtp_from_addr

    if not smtp_host and not notifications_conf.get("twilio_account_sid", None):
        logger.info("No SMTP nor SMS service configured, " "notification not sent")
        return

    # Worker in charge of sending notifications
    worker_session = Session(bind=worker_engine(app.config.repository))

    check = (
        worker_session.query(Check)
        .filter(Check.check_id == check_id)
        .join(Instance)
        .join(Host)
        .one()
    )

    port = check.instance.port
    hostname = check.instance.host.hostname
    instance = (
        worker_session.query(coreorm.Instance)
        .filter(coreorm.Instance.pg_port == port, coreorm.Instance.hostname == hostname)
        .one()
    )

    # don't notify if notifications are disabled for this instance
    if not instance.notify:
        return

    specs = check_specs
    spec = specs.get(check.name)

    message = ""
    if state != "OK":
        message = spec.get("message").format(
            key=key,
            check=check.name,
            value=value,
            threshold=getattr(check, state.lower()),
        )

    description = spec.get("description")
    subject = "[temBoard] {state} {hostname} - {description}".format(
        hostname=hostname, state=state, description=description
    )
    link = "https://%s:%d/server/%s/%d/alerting/%s" % (
        app.config.temboard.address,
        app.config.temboard.port,
        instance.agent_address,
        instance.agent_port,
        check.name,
    )

    direction = "➚" if prev_state == "OK" or state == "CRITICAL" else "➘"

    body = """
    Instance: {hostname}:{port}
    Description: {description}
    Status: {direction} {state} (prev. {prev_state})
    {message}
    {link}
    """.format(
        hostname=hostname,
        port=instance.agent_port,
        description=description,
        direction=direction,
        state=state,
        prev_state=prev_state,
        message=message,
        link=link,
    )

    notify = worker_session.execute(
        instance.select_email_and_phone_for_notify()
    ).fetchone()
    if len(notify.emails):
        send_mail(
            smtp_host,
            smtp_port,
            subject,
            body,
            notify.emails,
            smtp_tls,
            smtp_login,
            smtp_password,
            smtp_from_addr,
        )

    if len(notify.phones):
        send_sms(app.config.notifications, body, notify.phones)


# From itertools documentation
def grouper(n, iterable, fillvalue=None):
    """grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"""
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


@workers.schedule(id="schedule_collector", redo_interval=60)  # Repeat each 60s
@workers.register(pool_size=1)
def schedule_collector(app):
    """Worker function in charge of scheduling collector (pull mode)."""

    logger.setLevel(app.config.logging.level)
    logger.debug("Scheduling collect batches.")

    engine = worker_engine(app.config.repository)
    with engine.connect() as conn:
        # Get the list of agents
        res = conn.execute(
            "SELECT agent_address, agent_port "
            "FROM application.instances ORDER BY 1, 2"
        )

        for batch in grouper(16, res.fetchall()):
            batch = [row.values() for row in batch if row]

            logger.info("Scheduling batch collector for %s agents.", len(batch))

            collector_batch.defer(app, batch=batch)

    logger.debug("End of collector scheduler.")


@workers.register(pool_size=20)
def collector_batch(app, batch):
    # Start new ORM DB session
    engine = worker_engine(app.config.repository)
    engine.connect().close()  # Warm pool.

    for address, port in batch:
        try:
            collector(app, address, port, engine=engine)
        except UserError:
            raise
        except Exception as e:
            logger.exception("Failed to collect %s:%s: %s", address, port, e)


@workers.register(pool_size=20)
def collector(app, address, port, engine=None):
    agent_id = f"{address}:{port}"
    logger.info("Collecting metrics. agent=%s", agent_id)

    client = TemboardAgentClient.factory(app.config, address, port)
    # Start new ORM DB session
    engine = engine or worker_engine(app.config.repository)
    worker_session = Session(bind=engine)

    instance = get_instance(worker_session, address, port)
    worker_session.expunge(instance)
    host_id = instance_id = None
    # Agent monitoring API endpoint
    history_url = "/monitoring/history?limit=100"
    try:
        # Trying to find host_id, instance_id and the datetime of the latest
        # inserted record.

        # Find host_id and instance_id by hostname and PG port
        host_id = get_host_id(worker_session, instance.hostname)
        instance_id = get_instance_id(worker_session, host_id, instance.pg_port)
        logger.debug(
            "Found host #%s and instance #%s %s.", host_id, instance_id, instance
        )
    except Exception:
        # This case happens on the very first pull when no data have been
        # previously added.
        logger.debug(
            "Could not find host or instance records in monitoring inventory "
            "tables for %s.",
            instance,
        )
    else:
        # Get last inserted data timestamp from collector status
        collector_status = (
            worker_session.query(CollectorStatus)
            .filter(CollectorStatus.instance_id == instance_id)
            .first()
        )

        if collector_status and collector_status.last_insert:
            start = (collector_status.last_insert + timedelta(seconds=1)).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )
            history_url += "&start=%s" % start

    # Finally, let's call /monitoring/history agent API for getting metrics
    # history.
    try:
        logger.debug(
            "Querying monitoring history. instance=%s agent=%s", instance, agent_id
        )
        response = client.get(history_url)
        response.raise_for_status()
    except (OSError, ConnectionError, client.Error) as e:
        logger.error(
            "Failed to query agent: %s. Is it running? instance=%s agent=%s",
            e,
            instance,
            agent_id,
        )
        # Update collector status only if instance_id is known
        if instance_id:
            update_collector_status(
                worker_session, instance_id, "FAIL", last_pull=datetime.utcnow()
            )
            worker_session.commit()
        worker_session.close()
        return
    else:
        rows = response.json()

    # monitoring is still the better place to queue a discover. This allow us
    # to have sub-minute reactivity on instance change.
    discover_etag = response.headers.get("X-TemBoard-Discover-ETag")
    if discover_etag:
        if discover_etag != instance.discover_etag:
            logger.info("Detected discover data changes.")
            logger.debug(
                "New etag. old=%s new=%s", instance.discover_etag, discover_etag
            )
            if app.scheduler.can_schedule:
                refresh_discover.defer(app, address=address, port=port)
    else:
        logger.debug("Agent did not send discover ETag.")

    if not rows:
        logger.debug("Instance %s returned no monitoring data.", instance)

    for row in rows:
        logger.debug("Got points for %s at %s.", instance, row["datetime"])
        hostinfo = row["hostinfo"]
        data = row["data"]
        instance_d = row["instances"][0]

        try:
            # Try to insert collected data

            logger.debug("Update the inventory for %s. agent=%s", instance, agent_id)
            host = merge_agent_info(worker_session, hostinfo, instance_d)
            instance_id = get_instance_id(
                worker_session, host.host_id, instance.pg_port
            )
            logger.debug(
                "Insert instance availability for %s. agent=%s", instance, agent_id
            )
            insert_availability(
                worker_session, row["datetime"], instance_id, instance_d["available"]
            )
            worker_session.commit()
            logger.debug(
                "Insert collected metrics for %s. agent=%s", instance, agent_id
            )
            insert_metrics(
                worker_session,
                host.host_id,
                instance_id,
                data,
                dict(
                    agent=agent_id,
                    timestamp=(
                        # transform to ISOFORMAT (same as journalctl)
                        row["datetime"].replace(" +", "+").replace(" ", "T")
                    ),
                ),
                max_duration=app.config.monitoring.collect_max_duration,
            )
        except UserError:
            raise
        except DataError as e:
            # Wrong data type or corrupted data could lead to DataError. In
            # this case we should consider this row as unvalid and move to the
            # next one.
            try:
                last_insert = datetime.strptime(
                    row["datetime"], "%Y-%m-%d %H:%M:%S +0000"
                )
            except ValueError:
                # If row datetime could not be parsed, we should fallback to
                # current datetime. This will result to ignore potential valid
                # rows between row datetime and now, but this is better than
                # letting the collector stucked for ever on an invalid row.
                last_insert = datetime.utcnow()

            logger.exception(str(e))
            worker_session.rollback()
            if instance_id:
                logger.debug("Set collector status to FAIL for %s.", host)
                update_collector_status(
                    worker_session,
                    instance_id,
                    "FAIL",
                    last_pull=datetime.utcnow(),
                    last_insert=last_insert,
                )
                worker_session.commit()
            logger.debug("Continue with the next row.")
            continue

        logger.debug("Update collector status for agent %s.", agent_id)
        update_collector_status(
            worker_session,
            instance_id,
            "OK",
            last_pull=datetime.utcnow(),
            # This is the datetime format used by the agent
            last_insert=datetime.strptime(row["datetime"], "%Y-%m-%d %H:%M:%S +0000"),
        )
        worker_session.commit()
        logger.debug("Populate checks for %s.", host)
        # ALERTING PART
        populate_host_checks(
            worker_session, host.host_id, instance_id, dict(n_cpu=hostinfo["cpu_count"])
        )
        worker_session.commit()

        logger.debug(
            "Apply alerting checks against preprocessed data for agent %s.", agent_id
        )

        # Hack: copy max_connections in metrics data to pass max_connections to
        # alert processing. max_connections does NOT have the same type of
        # other metrics in data. This is because alerting.PreProcess functions
        # only accepts data as parameter.
        if "max_connections" in instance_d:
            row["data"]["max_connections"] = instance_d["max_connections"]

        try:
            check_preprocessed_data(
                app,
                worker_session,
                host.host_id,
                instance_id,
                preprocess_data(
                    row["data"],
                    get_instance_checks(worker_session, instance_id),
                    row["datetime"],
                ),
            )
        except Exception:
            logger.exception("Failed to check monitoring data for alerting.")

        logger.debug("Row with datetime=%s inserted", row["datetime"])
        worker_session.commit()

    worker_session.close()
    logger.debug("Collect done. agent=%s", agent_id)



import logging
from datetime import datetime, timedelta

from dateutil import parser as parse_datetime
from sqlalchemy.orm.exc import NoResultFound

from ...toolkit.errors import UserError
from ...web.tornado import HTTPError
from .alerting import bootstrap_checks, check_specs
from .model import db
from .model.orm import Check, CheckState, CollectorStatus, Host, Instance

logger = logging.getLogger(__package__)


def merge_agent_info(session, host_info, instance_info):
    """Update the host, instance and database information with the
    data received from the agent."""

    try:
        # Try to get host_id, based on hostname
        host_info["host_id"] = get_host_id(session, host_info["hostname"])
    except Exception:
        # host not found
        pass

    host = Host.from_dict(host_info)

    # Insert or update host information
    session.merge(host)
    session.flush()
    session.commit()

    # Get host_id in any case
    host_id = get_host_id(session, host_info["hostname"])
    # Weird that SA does not update this.
    host.host_id = host_id

    # Only process instances marked as available, since only those
    # have complete information
    if instance_info["available"]:
        try:
            # Try to get instance_id
            instance_info["instance_id"] = get_instance_id(
                session, host_id, instance_info["port"]
            )
        except Exception:
            # instance not found
            pass
        instance_info["host_id"] = host_id

        inst = Instance.from_dict(instance_info)
        # Insert or update instance information
        session.merge(inst)
        session.flush()
        session.commit()
    return host


def get_host_id(session, hostname):
    """Get host_id by the hostname."""
    host_id = db.get_host_id(session, hostname)
    if not host_id:
        raise NameError("Could not find registered host with hostname=%s" % hostname)
    return host_id


def get_instance_id(session, host_id, port):
    """Get instance_id by host_id and port."""
    instance_id = db.get_instance_id(session, host_id, port)
    if not instance_id:
        raise Exception(
            "Could not find registered instance with host_id=%s and port=%s"
            % (host_id, port)
        )
    return instance_id


def get_request_ids(request):
    host_id = get_host_id(request.db_session, request.instance.hostname)
    instance_id = get_instance_id(request.db_session, host_id, request.instance.pg_port)
    return host_id, instance_id


def parse_start_end(request):
    start = request.handler.get_argument("start", default=None)
    end = request.handler.get_argument("end", default=None)
    try:
        if start:
            start = parse_datetime.parse(start)
        if end:
            end = parse_datetime.parse(end)
    except ValueError:
        raise HTTPError(406, "Datetime not valid.")

    return start, end


class TimeoutError(UserError):
    pass


def insert_metrics(session, host_id, instance_id, data, labels=None, max_duration=30):
    start = datetime.utcnow()
    max_duration = timedelta(seconds=max_duration)
    labels = labels or {}

    for metric_name in list(data.keys()):
        call_duration = datetime.utcnow() - start
        if call_duration >= max_duration:
            logger.warning(
                "Metrics insertion is too long. "
                "Maybe another task is locking tables."
            )
            logger.warning(
                "Aborting metrics insertion. Retrying in less than a minute."
            )
            raise TimeoutError("Metrics insertion takes more than %s." % max_duration)

        # Do not try to insert empty lines
        if data[metric_name] is None:
            continue
        if len(data[metric_name]) == 0:
            continue

        for record in generate_logfmt_records(metric_name, data[metric_name]):
            try:
                logger.debug(
                    "up=1 %s %s",
                    " ".join(["%s=%s" % i for i in labels.items()]),
                    " ".join(["%s=%s" % i for i in record.items()]),
                )
            except Exception:
                logger.exception("Failed to format logfmt.")

        # Insert data
        for metric_data in data[metric_name]:
            logger.debug("Inserting data for metric %s.", metric_name)
            if metric_name == "sessions":
                db.insert_metric_sessions(session, instance_id, metric_data)
            elif metric_name == "xacts":
                db.insert_metric_xacts(session, instance_id, metric_data)
            elif metric_name == "locks":
                db.insert_metric_locks(session, instance_id, metric_data)
            elif metric_name == "blocks":
                db.insert_metric_blocks(session, instance_id, metric_data)
            elif metric_name == "bgwriter":
                db.insert_metric_bgwriter(session, instance_id, metric_data)
            elif metric_name == "db_size":
                db.insert_metric_db_size(session, instance_id, metric_data)
            elif metric_name == "tblspc_size":
                db.insert_metric_tblspc_size(session, instance_id, metric_data)
            elif metric_name == "filesystems_size":
                db.insert_metric_fs_size(session, host_id, metric_data)
            elif metric_name == "temp_files_size_delta":
                db.insert_metric_temp_files_size_delta(
                    session, instance_id, metric_data
                )
            elif metric_name == "wal_files":
                db.insert_metric_wal_files(session, instance_id, metric_data)
            elif metric_name == "cpu":
                db.insert_metric_cpu(session, host_id, metric_data)
            elif metric_name == "process":
                db.insert_metric_process(session, host_id, metric_data)
            elif metric_name == "memory":
                db.insert_metric_memory(session, host_id, metric_data)
            elif metric_name == "loadavg":
                db.insert_metric_loadavg(session, host_id, metric_data)
            elif metric_name == "vacuum_analyze":
                db.insert_metric_vacuum_analyze(session, instance_id, metric_data)
            elif metric_name == "replication_lag":
                db.insert_metric_replication_lag(session, instance_id, metric_data)
            elif metric_name == "replication_connection":
                db.insert_metric_replication_connection(
                    session, instance_id, metric_data
                )
            elif metric_name == "heap_bloat":
                db.insert_metric_heap_bloat(session, instance_id, metric_data)
            elif metric_name == "btree_bloat":
                db.insert_metric_btree_bloat(session, instance_id, metric_data)

            session.commit()


def generate_logfmt_records(metric, points):
    # Generate flat key-value record from temBoard metrics data for logfmt.
    if metric not in ("xacts",):
        return
    for point in points:
        record = dict()
        for k, v in point.items():
            if k in ("datetime", "port", "measure_interval"):
                continue
            if k in ("dbname",):
                logkey = k
            else:
                logkey = f"{metric}_{k}"
            record[logkey] = v
        yield record


def get_instance_checks(session, instance_id):
    # Returns enabled alerting checks as list of tuples:
    # (name, warning threshold, critical threshold)
    checks = session.query(Check).filter(Check.instance_id == instance_id)
    return [(c.name, c.warning, c.critical) for c in checks if c.enabled]


def populate_host_checks(session, host_id, instance_id, hostinfo):
    # Populate checks table with bootstraped checks if needed
    for bc in bootstrap_checks(hostinfo):
        # Do not try to add new check if exists
        if (
            session.query(Check)
            .filter(
                Check.host_id == host_id,
                Check.instance_id == instance_id,
                Check.name == bc[0],
            )
            .count()
            > 0
        ):
            continue
        c = Check(
            host_id=host_id,
            instance_id=instance_id,
            name=bc[0],
            enabled=True,
            warning=bc[1],
            critical=bc[2],
            description=check_specs.get(bc[0], {}).get("description"),
        )
        session.add(c)
    session.commit()


def update_collector_status(
    session, instance_id, status, last_pull=None, last_push=None, last_insert=None
):
    cs = CollectorStatus()
    cs.instance_id = instance_id
    cs.status = status
    if last_pull:
        cs.last_pull = last_pull
    if last_push:
        cs.last_push = last_push
    if last_insert:
        cs.last_insert = last_insert

    session.merge(cs)


def build_check_task_options(data, host_id, instance_id, checks, timestamp):
    """Build Task options for check_data_worker worker."""

    return dict(
        host_id=host_id,
        instance_id=instance_id,
        data=preprocess_data(data, checks, timestamp),
    )


def preprocess_data(data, checks, timestamp):
    """Preprocess metrics for alerting."""
    ret = list()

    for check in checks:
        spec = check_specs.get(check[0])
        if not spec:
            continue

        try:
            res = spec.get("preprocess")(data)
        except Exception as e:
            logger.debug("Failed to preprocess alerting check '%s': %s", check[0], e)
            continue

        if not isinstance(res, dict):
            res = {"": res}

        for key, value in list(res.items()):
            ret.append(
                dict(
                    datetime=timestamp,
                    name=check[0],
                    key=key,
                    value=value,
                    warning=check[1],
                    critical=check[2],
                )
            )
    return ret


def check_preprocessed_data(app, session, host_id, instance_id, ppdata):
    # Function in charge of checking preprocessed monitoring values
    keys = dict()

    for raw in ppdata:
        dt = raw.get("datetime")
        name = raw.get("name")
        key = raw.get("key")
        value = raw.get("value")
        warning = raw.get("warning")
        critical = raw.get("critical")

        # Proceed with thresholds comparison
        spec = check_specs.get(name)
        state = "UNDEF"
        if not spec:
            continue
        if not (
            spec.get("operator")(value, warning)
            or spec.get("operator")(value, critical)
        ):
            state = "OK"
        if spec.get("operator")(value, warning):
            state = "WARNING"
        if spec.get("operator")(value, critical):
            state = "CRITICAL"

        # Try to find enabled check for this host_id with the same name
        try:
            c = (
                session.query(Check)
                .filter(
                    Check.name == str(name),
                    Check.host_id == host_id,
                    Check.instance_id == instance_id,
                    Check.enabled == bool(True),
                )
                .one()
            )
        except NoResultFound:
            continue

        # Update/insert check current state
        try:
            cs = (
                session.query(CheckState)
                .filter(CheckState.check_id == c.check_id, CheckState.key == str(key))
                .one()
            )
            # State has changed since last time
            if cs.state != state:
                if app.scheduler.can_schedule:
                    app.scheduler.schedule_task(
                        "notify_state_change",
                        options={
                            "check_id": c.check_id,
                            "key": key,
                            "value": value,
                            "state": state,
                            "prev_state": cs.state,
                        },
                        expire=0,
                    )
                else:
                    logger.warning("Can't schedule state change task.")
            cs.state = str(state)
            session.merge(cs)

        except NoResultFound:
            cs = CheckState(check_id=c.check_id, key=str(key), state=str(state))
            session.add(cs)

        session.flush()
        # Append state change if any to history
        db.append_state_changes(
            session, dt, c.check_id, cs.state, cs.key, value, warning, critical
        )

        if c.check_id not in keys:
            keys[c.check_id] = list()
        keys[c.check_id].append(cs.key)

        session.commit()
        session.expunge_all()

    # Purge CheckState
    for check_id, ks in list(keys.items()):
        db.purge_check_states(session, check_id, ks)
        session.commit()

    # Get the list of check_id for the given instance
    req = session.query(Check).filter(Check.instance_id == instance_id)
    all_check_ids = [check.check_id for check in req]

    # Set to UNDEF each unchecked check for the given instance
    # This may happen when postgres is not available
    db.undef_check_states(session, all_check_ids, list(keys.keys()))
    session.commit()


# Stolen from ldap2pg
class Stopwatch:
    def __init__(self):
        self.delta = timedelta()

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.delta}>"

    def __enter__(self):
        self.start = datetime.utcnow()

    def __exit__(self, *_):
        self.last_delta = datetime.utcnow() - self.start
        self.delta += self.last_delta
        self.start = None



#
# There is three app objects in temBoard. One for Tornado, Flask and temBoard
# itself.
#
# Tornado's app is implemented and instanciated in t.w.tornado. Tornado's app
# holds Tornado routes. Tornado's app is a Tornado HTTP server delegate.
#
# Flask app is implemented in t.w.flask and instanciated here in t.__main__,
# accessible through flask.current_app thread local. Flask app holds Flask
# routes. Flask app is a WSGI app.
#
# temBoard app is implemented and instanciated in t.c.app. temBoard app holds
# configuration, CLI definition and controls execution of the service. temBoard
# app is a CLI entrypoint.
#
# Each of this app has different purpose, configuration handling and
# initialization process. This is quite messy and needs redesign.
#

import sys

from .cli.app import app
from .model import QUERIES
from .web.flask import create_app


def main(*a, **kw):
    flask = create_app(app)
    flask.app_context().push()

    # Import main routes now current_app is up.
    __import__(__package__ + ".web.routes.core")
    __import__(__package__ + ".web.routes.auth")
    __import__(__package__ + ".web.routes.inventory")
    __import__(__package__ + ".web.routes.home")
    __import__(__package__ + ".web.routes.instance")
    __import__(__package__ + ".web.routes.settings")

    # Load sub commands.
    __import__(__package__ + ".cli.apikey")
    __import__(__package__ + ".cli.generate_key")
    # Don't import prometheus command yet.
    # Use python -m temboardui.cli.prometheus until 9.0.
    __import__(__package__ + ".cli.migratedb")
    __import__(__package__ + ".cli.query_agent")
    __import__(__package__ + ".cli.register_instance")
    __import__(__package__ + ".cli.routes")
    __import__(__package__ + ".cli.serve")
    __import__(__package__ + ".cli.tasks")
    __import__(__package__ + ".cli.web")

    QUERIES.load()

    return app(*a, **kw)


if "__main__" == __name__:
    sys.exit(main())



from ...web.flask import instance_proxy


class DashboardPlugin:
    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        __import__(__name__ + ".routes")
        instance_proxy.generic_proxy("/dashboard")



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import backref, relationship

from temboardui.plugins.monitoring.model import tables


class ModelMixin:
    """
    Class to be mixed in with every mapped class
    """

    @classmethod
    def from_dict(cls, dict_values, recurse=False):
        """
        Build an instance of this model from a dictionary.
        Additional keys are discarded along the way.

        Args:

        dict_values (dict):
            a dictionary from which the instance should be built
        recurse (bool):
            if ``True``, also builds related instances. Else, only the base
            columns are included.
        """
        mapper = inspect(cls)
        columns_values = {
            key: value
            for key, value in list(dict_values.items())
            if key in mapper.column_attrs
        }
        if recurse:
            items = mapper.relationships.items()
            for key, relationship_property in list(items):
                target_cls = relationship_property.mapper.class_
                value = dict_values.get(key)
                if isinstance(value, dict):
                    columns_values[key] = target_cls.from_dict(value)
                elif isinstance(value, list):
                    subvalues = columns_values.setdefault(key, [])
                    for val in value:
                        if isinstance(val, dict):
                            subvalues.append(target_cls.from_dict(val))
                        else:
                            subvalues.append(target_cls(val))
        return cls(**columns_values)


Model = declarative_base(cls=(ModelMixin,))


class Host(Model):
    __table__ = tables.hosts

    def __repr__(self):
        return "host: %s" % self.hostname


class Instance(Model):
    __table__ = tables.instances
    host = relationship("Host", backref="instances")

    def __repr__(self):
        return f"instance: {self.host_id}:{self.port}"


class Check(Model):
    __table__ = tables.checks
    host = relationship("Host", backref="checks")
    instance = relationship("Instance", backref="checks")


class CheckState(Model):
    __table__ = tables.checkstates
    check = relationship("Check", backref="states")


class CollectorStatus(Model):
    __table__ = tables.collector_status
    instance = relationship(
        "Instance", backref=backref("collector_statuts", cascade="all, delete-orphan")
    )



from ...web.flask import instance_proxy


class MaintenancePlugin:
    def __init__(self, app):
        self.app = app

    def load(self):
        __import__(__name__ + ".routes")
        instance_proxy.generic_proxy("/maintenance")
        instance_proxy.generic_proxy("/maintenance/<database>")
        instance_proxy.generic_proxy("/maintenance/<database>", method="POST")
        instance_proxy.generic_proxy(
            "/maintenance/<any(vacuum,analyze,reindex):action>/<database>",
            method="DELETE",
        )
        instance_proxy.generic_proxy(
            "/maintenance/<database>/<any(vacuum,analyze,reindex):action>",
            method="POST",
        )
        instance_proxy.generic_proxy(
            "/maintenance/<database>/<any(vacuum,analyze,reindex):action>/scheduled"
        )
        instance_proxy.generic_proxy("/maintenance/<database>/schema/<schema>")
        instance_proxy.generic_proxy(
            "/maintenance/<database>/schema/<schema>/<any(vacuum,analyze,reindex):action>/scheduled"
        )
        instance_proxy.generic_proxy(
            "/maintenance/<database>/schema/<schema>/table/<table>"
        )
        instance_proxy.generic_proxy(
            "/maintenance/<database>/schema/<schema>/<any(index,table):type>/<object>/reindex",
            method="POST",
        )
        instance_proxy.generic_proxy(
            "/maintenance/<database>/schema/<schema>/table/<table>/<any(vacuum,analyze):action>",
            method="POST",
        )
        instance_proxy.generic_proxy(
            "/maintenance/<database>/schema/<schema>/table/<table>/<any(vacuum,analyze):action>/scheduled"
        )



# Flask WSGI app is served by Tornado's fallback handler.

import contextlib
import functools
import json
import logging
import os
from ipaddress import ip_address, ip_network

import jinja2
import werkzeug.exceptions
from flask import (
    Blueprint,
    Flask,
    abort,
    current_app,
    g,
    jsonify,
    make_response,
    render_template,
    request,
)
from tornado.web import decode_signed_value
from werkzeug.exceptions import HTTPException

from ..agentclient import TemboardAgentClient
from ..application import get_instance, get_role_by_cookie
from ..model import Session
from ..model.orm import ApiKey, StubRole
from ..toolkit.utils import utcnow
from .tornado import serialize_querystring
from .vitejs import ViteJSExtension

logger = logging.getLogger(__name__)


class InstanceProxyBlueprint(Blueprint):
    # Pass-through implementation for /proxy/address/port/…
    def generic_proxy(self, url, method="GET"):
        # Register an unique endpoint name for each proxy.
        # Without this, every generic proxy routes would share the same
        # endpoint, leading to :
        # View function mapping is overwriting an existing endpoint
        endpoint = f"generic_instance_proxy_{url.replace('/', '_')}_{method}"

        @self.route(url, methods=[method], endpoint=endpoint)
        def generic_instance_proxy(**_):
            # Skip the first three segments proxy/adress/port
            *_, path = request.path.split("/", 4)
            path = f"/{path}"

            response = current_app.instance.request(
                path,
                method=method,
                body=request.get_json() if method == "POST" else None,
                query=request.args.to_dict(),
            )
            return jsonify(response.json())


# InstanceMiddleware extension controls request context for the following
# blueprint routes.
instance_proxy = InstanceProxyBlueprint(
    "instance_proxy", __name__, url_prefix="/proxy/<address>/<port>"
)

instance_routes = Blueprint(
    "instance_routes", __name__, url_prefix="/server/<address>/<port>"
)


def create_app(temboard_app):
    app = Flask("temboardui")
    app.start_time = utcnow()
    app.config["DEBUG"] = app.config["TESTING"] = "DEBUG" in os.environ
    app.temboard = temboard_app
    app.static_folder = "static/dist"
    app.static_url_path = "/static"
    app.template_folder = "templates/flask"
    app.jinja_env.undefined = jinja2.StrictUndefined
    app.jinja_env.trim_blocks = True
    SQLAlchemy(app)
    APIKeyMiddleware(app)
    UserMiddleware(app)
    AuthMiddleware(app)
    # DEPRECATED: Seems required for old Flask only.
    for code in werkzeug.exceptions.default_exceptions.keys():
        app.register_error_handler(code, error_handler)
    app.register_error_handler(Exception, error_handler)

    # unsafe-eval is for jquery. unsafe-inline because we have
    # script tags in templates.
    csp = "default-src 'self' 'unsafe-inline' 'unsafe-eval' data:"
    if "VITEJS" in os.environ:
        csp += " localhost:5173 ws:"

    @app.after_request
    def add_csp(resp):
        resp.headers["Content-Security-Policy"] = csp
        return resp

    ViteJSExtension(app)
    # finalize_app() must be called before serving, to enable configured
    # blueprints.
    return app


def finalize_app():
    # Configure Flask app after configuration is loaded.
    app = current_app

    # This middleware registers instance_proxy blueprint, loads g.instance
    # object and provides helpers in app.instance. instance is loaded only for
    # instance_proxy blueprint. instance_proxy must be registered after plugins
    # loading.
    InstanceMiddleware(app)

    return app


def error_handler(e):
    status_code = e.code if isinstance(e, HTTPException) else None

    error = str(e) or repr(e)
    if hasattr(e, "description"):  # From Flask HTTPException.
        error = e.description

    if status_code is None:
        logger.exception("Unhandled error:")
        status_code = 500
        if current_app.temboard.debug:
            error = "Internal temBoard error. Retry later or contact administrator."
    elif status_code < 500:
        logger.warning("User error: %s", error)
    else:
        logger.error("Fatal error: %s", error)

    if request.path.endswith(".csv"):
        return "", status_code
    elif is_json(request.path):
        response = jsonify(error=error)
        response.status_code = status_code
        return response

    template_vars = {}
    if hasattr(g, "instance"):
        template_vars["instance"] = g.instance
    if status_code in (401, 403, 404):
        template = f"{status_code}.html"
    else:
        template = "error.html"

    return render_template(template, message=error, code=status_code, **template_vars)


def is_json(path):
    return (
        path.startswith("/json") or path.startswith("/proxy") or path.endswith(".json")
    )


class SQLAlchemy:
    # Flask extension to manage a SQLAlchemy session per request.
    def __init__(self, app=None):
        self.app = app
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.db = self
        app.before_request(self.before)
        app.teardown_request(self.teardown)

    def before(self):
        g.db_session = Session()

    def teardown(self, error):
        # Expunge objects to implement expire_on_rollback=False.
        # This allow templates to reuse request.instance object and joined object
        # without triggering lazy load.
        g.db_session.expunge_all()
        g.db_session.close()
        del g.db_session


class APIKeyMiddleware:
    # Flask extension validating API key header

    def __init__(self, app):
        self.app = app
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.apikey = self
        app.before_request(self.before)

    def before(self):
        g.apikey = None

        if "Authorization" not in request.headers:
            return

        remote_addr = ip_address(request.remote_addr)
        if not any(
            remote_addr in ip_network(cidr)
            for cidr in self.app.temboard.config.auth.allowed_ip
        ):
            logger.debug("Authorization ignored for IP %s.", remote_addr)
            return

        try:
            scheme, secret = request.headers["Authorization"].split(None, 1)
        except TypeError:
            abort(400, "Malformed Authorization header")

        if scheme != "Bearer":
            logger.debug("Ignoring Authorization scheme %s.", scheme)
            return

        key = ApiKey.select_secret(secret).with_session(g.db_session).scalar()
        if not key:
            abort(403, "Unknown API Key.")

        if key.expired:
            abort(403, "Expired API key.")

        logger.debug("Accepted API key from HTTP Header, client IP %s.", remote_addr)
        g.apikey = key


class AuthMiddleware:
    # Flask extension enforcing authentication

    def __init__(self, app=None):
        self.app = app
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.user = self
        app.before_request(self.before)

        @app.context_processor
        def inject_user():
            return dict(role=g.current_user)

    def before(self):
        if request.url_rule and request.url_rule.rule.startswith("/static"):
            # Skip Auth for static files.
            return

        func = self.app.view_functions.get(request.endpoint)
        if not func:
            abort(404)

        apikey_allowed = getattr(func, "__apikey_allowed", False)
        if apikey_allowed and g.apikey:
            logger.debug("Endpoint authorized by API key.")
            return

        anonymous_allowed = getattr(func, "__anonymous_allowed", False)
        if not anonymous_allowed and g.current_user is None:
            logger.debug("Refusing anonymous access.")
            abort(401)

        admin_required = getattr(func, "__admin_required", False)
        if admin_required and not g.current_user.is_admin:
            logger.debug("Refusing access to non-admin user.")
            abort(403)


class UserMiddleware:
    # Flask extension to load current user.

    def __init__(self, app=None):
        self.app = app
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.user = self
        app.before_request(self.before)

    def before(self):
        g.current_user = None
        if "temboard" in request.cookies:
            g.current_user = self.load_user_from_tornado_secure_cookie()
        if getattr(g, "apikey", None):
            g.current_user = StubRole("temboard")

    def load_user_from_tornado_secure_cookie(self):
        cookie = decode_signed_value(
            self.app.temboard.config.temboard.cookie_secret,
            "temboard",
            request.cookies["temboard"],
        )
        if cookie:
            cookie = cookie.decode("utf-8")
            return get_role_by_cookie(g.db_session, cookie)


class InstanceMiddleware:
    # Flask extension providing instance helpers to view, setting up
    # instance_proxy middleware. Must be initialize once instance_proxy has its
    # routes.

    def __init__(self, app=None):
        self.app = app
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.instance = self
        app.register_blueprint(instance_proxy)
        app.register_blueprint(instance_routes)
        brf = app.before_request_funcs
        brf[instance_proxy.name] = [self.load_instance_before_request]
        brf[instance_routes.name] = [self.load_instance_before_request]

        @app.teardown_request
        def teardown_instance(*_):
            if hasattr(g, "instance"):
                delattr(g, "instance")

        @app.context_processor
        def context_processor():
            if not hasattr(g, "instance"):
                return {}
            return dict(instance=g.instance)

    def load_instance_before_request(self):
        address = request.view_args.pop("address")
        port = request.view_args.pop("port")

        g.instance = get_instance(g.db_session, address, port)
        if not g.instance:
            abort(404)

        func = self.app.view_functions.get(request.endpoint)
        apikey_allowed = g.apikey and getattr(func, "__apikey_allowed", False)
        user_allowed = (
            g.current_user
            and g.db_session.execute(
                g.instance.has_dba(g.current_user.role_name)
            ).scalar()
        )
        if not apikey_allowed and not user_allowed:
            abort(403)
        g.instance.status = None
        prefix = current_app.blueprints[request.blueprint].url_prefix
        request.instance_path = request.url_rule.rule.replace(prefix, "")

    def fetch_status(self):
        try:
            data = json.load(self.request("/status"))
        except Exception as e:
            # agent is unreachable we forge a status response
            logger.error("Failed to fetch status: %s", e)
            data = {
                "temboard": {"status": "unreachable"},
                "postgres": {"status": "unreachable", "pending_restart": False},
                "system": {"status": "unreachable"},
            }
        g.instance.status = data

    def client(self):
        return TemboardAgentClient.factory(
            current_app.temboard.config,
            g.instance.agent_address,
            g.instance.agent_port,
            g.current_user.role_name,
        )

    def request(self, path, method="GET", query=None, body=None):
        client = self.client()

        pathinfo = path
        if query:
            pathinfo += "?" + serialize_querystring(query)

        try:
            response = client.request(method=method, path=pathinfo, body=body)
            response.raise_for_status()
        except ConnectionError as e:
            abort(500, str(e))
        except TemboardAgentClient.Error as e:
            abort(e.response.status, e.message)
        except Exception as e:
            logger.error("Proxied request failed: %s", e)
            abort(500)
        else:
            return response

    def proxy(self):
        # Translate current Flask request to instance agent, translate the
        # response to a Flask response ready for return.
        agent_response = self.request(path=request.instance_path, method=request.method)
        response = make_response(agent_response.read())
        h = "Content-Type"
        response.headers[h] = agent_response.headers[h]
        return response

    def check_active_plugin(self, name):
        """
        Ensure that the plugin is active for given instance
        """
        if name not in [p.plugin_name for p in g.instance.plugins]:
            raise abort(408, "Plugin %s not activated." % name)


def anonymous_allowed(func):
    # Decorator marking a route as public.
    func.__anonymous_allowed = True
    return func


def apikey_allowed(func):
    # Decorator allowing a route by apikey auth
    func.__apikey_allowed = True
    return func


def admin_required(func):
    # Similar to flask_security.roles_required, but limited to admin role.
    func.__admin_required = True
    return func


def transaction(func):
    # Flask is not reliable for catching exception in extension.
    # Instead, use an explicit decorator to deduplicate transaction handling.
    # Use this in routes modifying database.
    @functools.wraps(func)
    def autocommit_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            g.db_session.rollback()
            raise
        finally:
            g.db_session.commit()

    return autocommit_wrapper


@contextlib.contextmanager
def validating():
    # Translate ValueError to HTTP 408
    try:
        yield
    except ValueError as e:
        raise abort(408, str(e))



from ..model import check_schema
from ..toolkit import services
from ..toolkit.app import SubCommand
from .app import app


@app.command
class Serve(SubCommand):
    """Combined web server and background workers."""

    is_service = True

    def main(self, args):
        check_schema()
        self.app.config.load_signing_key()

        services.run(self.app.webservice, self.app.scheduler, self.app.worker_pool)



# Integration of ViteJS in temBoard
import json
import logging
import os
from textwrap import dedent

logger = logging.getLogger(__name__)


class ViteJSExtension:
    # Flask extension managing ViteJS

    def __init__(self, app=None):
        self.app = app
        self.manifest = None
        if app:
            self.init_app(app)

    @property
    def PROD(self):
        return not bool(self.app.config.get("VITEJS"))

    def init_app(self, app):
        app.vitejs = self
        app.config["VITEJS"] = uri = os.environ.get("VITEJS", "").rstrip("/")
        if uri and not uri.startswith("http://localhost:"):
            raise Exception("Refusing to delegate to non-localhost ViteJS.")
        self.manifest_path = self.app.static_folder + "/.vite/manifest.json"
        app.context_processor(self.context_processor)

    def read_manifest(self):
        if self.app.config["VITEJS"]:
            logger.debug("Using ViteJS dev server at %s.", self.app.config["VITEJS"])
            logger.debug("Skip reading ViteJS manifest.")
            return

        logger.debug("Loading ViteJS manifest.")
        with open(self.manifest_path) as fo:
            self.manifest = json.load(fo)

    def url_for(self, name):
        if self.manifest:
            return self.app.static_url_path + "/" + self.manifest[name]["file"]
        else:
            return self.app.config["VITEJS"] + "/static/" + name

    def css_links_for(self, name):
        if not self.manifest:
            return

        for css in self.manifest[name].get("css", []):
            yield self.tag_for(css)

        for import_ in self.manifest[name].get("imports", []):
            yield from self.css_links_for(import_)

    def tag_for(self, file_):
        if file_.endswith("css"):
            return dedent("""\
            <link rel="stylesheet" href="%s/%s" />
            """) % (self.app.static_url_path, file_)
        else:
            raise Exception("Unhandled assets file type for %s" % file_)

    def context_processor(self):
        return dict(vitejs=self)



from flask import current_app as app
from flask import g, jsonify, redirect

from ...model.orm import Instance
from ..flask import admin_required, anonymous_allowed


@app.route("/")
@anonymous_allowed
def index():
    if g.current_user:
        return redirect("/home")
    return redirect("/login")


@app.route("/json/instances/home")
def get_instance_home():
    """Data for InstanceCards.vue component."""
    return jsonify(
        [
            {k: getattr(row, k) for k in row.keys()}
            for row in g.db_session.execute(
                Instance.select_for_home(g.current_user.role_name)
            )
        ]
    )


@app.route("/json/plugins")
@admin_required
def get_plugins():
    """List plugins."""
    return jsonify(sorted(app.temboard.config.temboard.plugins))



from flask import abort, g, jsonify, render_template, request
from flask import current_app as app

from ...application import send_mail, send_sms
from ...model import orm
from ..flask import admin_required


@app.route("/settings/instances")
@admin_required
def settings_instances():
    return render_template(
        "settings/instances.html",
        sidebar=True,
        instance_list=orm.Instance.all().with_session(g.db_session).all(),
    )


@app.route("/settings/environments")
@admin_required
def settings_environments():
    return render_template(
        "settings/environments.html",
        sidebar=True,
        environments=orm.Environment.all().with_session(g.db_session).all(),
    )


@app.route("/settings/environments/<name>/members")
@admin_required
def settings_environment_members(name):
    return render_template("settings/members.html", sidebar=True, environment=name)


@app.route("/settings/users")
@admin_required
def settings_users():
    return render_template(
        "settings/users.html",
        sidebar=True,
        role_list=orm.Role.all().with_session(g.db_session).all(),
    )


@app.route("/settings/notifications")
@admin_required
def settings_notifications():
    return render_template(
        "settings/notifications.html",
        sidebar=True,
        email_configured=app.temboard.config.notifications.smtp_host,
        sms_configured=app.temboard.config.notifications.twilio_account_sid,
    )


@app.route("/json/test_email", methods=["POST"])
@admin_required
def post_test_email():
    email = request.json.get("email")
    if not email:
        abort(400, "Email field is missing.")

    notifications_conf = app.temboard.config.notifications

    if not notifications_conf.smtp_host:
        abort(406, "SMTP server is not configured")

    send_mail(
        notifications_conf.smtp_host,
        notifications_conf.smtp_port,
        "temBoard notification test",
        "This is a test",
        email,
        notifications_conf.smtp_tls,
        notifications_conf.smtp_login,
        notifications_conf.smtp_password,
        notifications_conf.smtp_from_addr,
    )
    return jsonify({"message": "OK"})


@app.route("/json/test_sms", methods=["POST"])
@admin_required
def post_test_sms():
    phone = request.json.get("phone")
    if not phone:
        abort(400, "Phone field is missing.")

    send_sms(app.temboard.config.notifications, "temBoard notification test", [phone])
    return jsonify({"message": "OK"})



"""
Mapping of the database for use with SQLAlchemy.
"""

from sqlalchemy.schema import Column, ForeignKey, MetaData, Table, UniqueConstraint
from sqlalchemy.sql import text
from sqlalchemy.types import REAL, BigInteger, Boolean, DateTime, Integer, UnicodeText

metadata = MetaData()

hosts = Table(
    "hosts",
    metadata,
    Column("host_id", Integer, primary_key=True),
    Column("hostname", UnicodeText, nullable=False, unique=True),
    Column("os", UnicodeText, nullable=False),
    Column("os_version", UnicodeText, nullable=False),
    Column("os_flavour", UnicodeText),
    Column("cpu_count", Integer),
    Column("cpu_arch", UnicodeText),
    Column("memory_size", BigInteger),
    Column("swap_size", BigInteger),
    Column("virtual", Boolean),
    schema="monitoring",
)

instances = Table(
    "instances",
    metadata,
    Column("instance_id", Integer, primary_key=True),
    Column("host_id", Integer, ForeignKey("monitoring.hosts.host_id"), nullable=False),
    Column("port", Integer, nullable=False),
    Column("local_name", UnicodeText, nullable=False),
    Column("version", UnicodeText, nullable=False),
    Column("version_num", Integer, nullable=False),
    Column("data_directory", UnicodeText, nullable=False),
    Column("sysuser", UnicodeText),
    Column("standby", Boolean, nullable=False, server_default=text("False")),
    UniqueConstraint("host_id", "port"),
    schema="monitoring",
)

checks = Table(
    "checks",
    metadata,
    Column("check_id", Integer, primary_key=True),
    Column("host_id", Integer, ForeignKey("monitoring.hosts.host_id"), nullable=False),
    Column(
        "instance_id",
        Integer,
        ForeignKey("monitoring.instances.instance_id"),
        nullable=True,
    ),
    Column("enabled", Boolean, nullable=False),
    Column("name", UnicodeText, nullable=False),
    Column("warning", REAL),
    Column("critical", REAL),
    Column("description", UnicodeText),
    schema="monitoring",
)

checkstates = Table(
    "check_states",
    metadata,
    Column(
        "check_id",
        Integer,
        ForeignKey("monitoring.checks.check_id"),
        nullable=False,
        primary_key=True,
    ),
    Column("key", UnicodeText, nullable=True, primary_key=True),
    Column("state", UnicodeText, nullable=False),
    schema="monitoring",
)

collector_status = Table(
    "collector_status",
    metadata,
    Column(
        "instance_id",
        Integer,
        ForeignKey("monitoring.instances.instance_id"),
        primary_key=True,
    ),
    Column("last_pull", DateTime, nullable=True),
    Column("last_push", DateTime, nullable=True),
    Column("last_insert", DateTime, nullable=True),
    Column("status", UnicodeText),
    schema="monitoring",
)



import logging
import sys
from time import sleep

from sqlalchemy import __version__ as sa_version
from sqlalchemy import create_engine
from sqlalchemy.orm.session import sessionmaker

from ..toolkit.queries import QueryFiler
from .migrator import Migrator

Session = sessionmaker(expire_on_commit=False)
logger = logging.getLogger(__name__)
# named queries, loaded with QUERIES.load() by temboardui.__main__.
QUERIES = QueryFiler(__path__[0] + "/queries")


def format_dsn(dsn):
    fmt = (
        "postgresql://{user}:{password}@:{port}"
        "/{dbname}"
        "?host={host}&application_name=temboard"
    )
    return fmt.format(**dsn)


def configure(dsn, **kwargs):
    if hasattr(dsn, "items"):
        dsn = format_dsn(dsn)

    try:
        engine = create_engine(dsn)
    except Exception as e:
        logger.warning("Connection to the database failed: %s", e)
        logger.warning("Please check your configuration.")
        sys.stderr.write("FATAL: %s\n" % e)
        exit(10)
    Session.configure(bind=engine, **kwargs)

    # For legacy purpose, we return engine to ease binding other Session maker
    # with the same engine.
    return engine


def check_connectivity(engine):
    for i in range(1, 10):
        try:
            # Use raw psycopg2 conn to workaround old SA bug in initialization.
            conn = engine.pool._invoke_creator()
            with conn.cursor() as cur:
                cur.execute("SELECT version();")
                (pgversion,) = cur.fetchone()
                logger.debug("Using PostgreSQL %s.", pgversion)
            conn.close()
            break
        except Exception as e:
            if i == 9:
                raise
            logger.warning("Failed to connect to database: %s", e)
            logger.info("Retrying in %ss.", i)
            sleep(i)

    server, pgversion, _ = pgversion.split(" ", 2)
    if "." not in pgversion and sa_version < "1.2":
        msg = "SQLAlchemy 1.1 does not support version %s."
        raise Exception(msg % pgversion)


def worker_engine(dbconf):
    """Create a new stand-alone SQLAlchemy engine to be instantiated in worker
    context.
    """
    return create_engine(format_dsn(dbconf))


def check_schema():
    engine = Session.kw["bind"]
    check_connectivity(engine)
    migrator = Migrator()
    migrator.inspect_available_versions()
    migrator.inspect_current_version(engine.raw_connection().connection)
    migrator.check()



import logging
import sys
from urllib.parse import urlparse

from ..agentclient import TemboardAgentClient
from ..toolkit import validators as v
from ..toolkit.app import SubCommand
from ..toolkit.errors import UserError
from .app import app

logger = logging.getLogger(__name__)


@app.command
class QueryAgent(SubCommand):
    """Query a temBoard agent

    Accepts body contents in stdin. temBoard agent accepts only JSON as body.

    """

    name = "query-agent"

    def define_arguments(self, parser):
        parser.add_argument(
            "--delete", action="store_const", const="DELETE", dest="method"
        )

        parser.add_argument("--post", action="store_const", const="POST", dest="method")

        parser.add_argument(
            "--username",
            metavar="USERNAME",
            default="temboard",
            help="temBoard UI Username",
        )

        parser.add_argument(
            "url", metavar="URL", help="Full URL to agent to query.", type=v.url
        )

        parser.add_argument(
            "body",
            default=None,
            nargs="?",
            metavar="JSON",
            help="JSON payload for POST request.",
        )

        super().define_arguments(parser)

    def main(self, args):
        url = urlparse(args.url)
        headers = {}

        if not args.body and not sys.stdin.isatty():
            logger.info("Reading request body from STDIN.")
            args.body = sys.stdin.read()

        if args.body:
            headers["Content-Type"] = "application/json"

        method = "POST" if args.body else getattr(args, "method", "GET")
        pathinfo = url.path
        if url.query:
            pathinfo = f"{pathinfo}?{url.query}"

        client = TemboardAgentClient.factory(
            self.app.config, url.hostname, url.port, username=args.username
        )
        client.log_headers = True

        try:
            response = client.request(method, pathinfo, headers, args.body)
            sys.stdout.write(response.read().decode("utf-8"))
            response.raise_for_status()
        except ConnectionError as e:
            logger.critical("%s", e)
            return 1
        except client.Error as e:
            raise UserError(str(e))

        return 0



from ..web.tornado import app, render_template


@app.instance_route(r"/notifications")
def notifications(request):
    notifications = request.instance.get("/notifications")
    request.instance.fetch_status()
    return render_template(
        "notifications.html",
        instance=request.instance,
        notifications=notifications,
        plugin="notifications",
        role=request.current_user,
    )



import logging
import os.path
from subprocess import check_call

from cryptography.hazmat.primitives import serialization

from ..toolkit.app import SubCommand
from ..toolkit.errors import UserError
from ..toolkit.signing import load_private_key
from .app import app

logger = logging.getLogger(__name__)


@app.command
class GenerateKey(SubCommand):
    """Generate signing key."""

    name = "generate-key"

    def define_arguments(self, parser):
        parser.add_argument(
            "--force",
            action="store_true",
            default=False,
            help="Force overwriting existing files.",
        )

    def main(self, args):
        priv = self.app.config.temboard.signing_private_key
        if os.path.exists(priv) and not args.force:
            raise UserError("%s exists. Use --force to overwrite." % priv)
        logger.info("Generating RSA key with openssl at %s.", priv)
        check_call(["openssl", "genrsa", "-out", priv, "4096"])

        with open(priv, "rb") as fo:
            privkey = load_private_key(fo.read())

        pub = self.app.config.temboard.signing_public_key
        logger.info("Exporting public key at %s.", pub)
        pem = privkey.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        with open(pub, "wb") as fo:
            fo.write(pem)

        return 0



#
# This module implements protocol switching from HTTP to HTTPS on the same
# socket connection.
#
# In tornado, a bunch of objects interacts with each other to handle a request:
# IOStream, TCP/HTTPServer, HTTP1Connection, HTTPMessageDelegate, HTTPRequest,
# HTTPResponse and HTTPRequestHandler. But it's not as simple as an onion of
# successive encapsulations.
#
# OpenSSL raises an error on uncrypted HTTP request while waiting for
# handshake. The first step is to have an IOStream with explicit SSL handshake
# allowing to catch this OpenSSL error. This stream must NOT close the socket
# on error.
#
# On HTTP request, we divert the request processing to the protocol switcher.
# The protocol switcher almost look like a regular request/response handling
# but there is no routing. We limit processing to headers to preserver host and
# path of the request.
#
# All of this (handshake, read request and write response) must be async to
# keep the server loop managing all incoming connections.

import errno
import logging
import socket
import ssl

from tornado import gen
from tornado.escape import native_str
from tornado.http1connection import HTTP1Connection
from tornado.httpclient import HTTPResponse
from tornado.httpserver import HTTPRequest, HTTPServer
from tornado.httputil import HTTPHeaders, ResponseStartLine, parse_request_start_line
from tornado.ioloop import IOLoop
from tornado.iostream import IOStream, SSLIOStream, StreamClosedError
from tornado.log import app_log, gen_log
from tornado.netutil import ssl_options_to_context
from tornado.util import errno_from_exception

logger = logging.getLogger(__name__)


def parse_http_headers(payload):
    payload = native_str(payload)
    # Implements simple HTTP1Connection._read_message but IO-free.
    lines = payload.splitlines()
    if lines and ":" not in lines[0]:
        # Drop start line
        lines.pop(0)
    # Drop contents
    if "" in lines:
        lines[:] = lines[: lines.index("")]
    return (
        parse_request_start_line("GET / HTTP/1.1"),
        HTTPHeaders.parse("\r\n".join(lines)),
    )


def protocol_switcher(request):
    try:
        host = request.headers["Host"]
    except KeyError:
        # We don't have FQDN. Fallback to socket address. This breaks
        # name-based virtualhost.
        host = "%(address)s:%(port)s" % dict(
            request.config.temboard, address=request.host
        )
    new_url = f"https://{host}{request.uri}"
    headers = HTTPHeaders({"Content-Length": "0", "Location": new_url})
    logger.debug("Redirecting client to %s.", new_url)
    return HTTPResponse(
        request=request,
        code=301,
        headers=headers,
        # If effective_url is not set, HTTPResponse falls back to request.url,
        # which does not exists... See tornado.httpclient.HTTPResponse.__init__
        # and tornado.httpserver.HTTPRequest.
        effective_url=request.full_url(),
    )


class SSLErrorHTTPRequest(Exception):
    pass


class EasySSLIOStream(SSLIOStream):
    # SSIOStream raising exception on HTTP_REQUEST rather than closing socket.

    def _do_ssl_handshake(self):
        if self._ssl_connect_future.done():
            # After HTTP_REQUEST is detected, we may still have read event.
            # Just ignore them.
            gen_log.debug("Ignoring subsequent read event.")
            return

        # Based on code from test_ssl.py in the python stdlib
        try:
            self._handshake_reading = False
            self._handshake_writing = False
            self.socket.do_handshake()
        except ssl.SSLError as err:
            if err.args[0] == ssl.SSL_ERROR_WANT_READ:
                self._handshake_reading = True
                return
            elif err.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                self._handshake_writing = True
                return
            elif err.args[0] in (ssl.SSL_ERROR_EOF, ssl.SSL_ERROR_ZERO_RETURN):
                return self.close(exc_info=True)
            elif err.args[0] == ssl.SSL_ERROR_SSL:
                try:
                    peer = self.socket.getpeername()
                except Exception:
                    peer = "(not connected)"

                if getattr(err, "reason", None) == "HTTP_REQUEST":
                    # Async raise HTTP_REQUEST error.
                    self._ssl_connect_future.set_exception(SSLErrorHTTPRequest())
                    gen_log.warning("HTTP_REQUEST on SSL handshake from %s.", peer)
                    return

                gen_log.warning(
                    "SSL Error on %s %s: %s", self.socket.fileno(), peer, err
                )
                return self.close(exc_info=True)
            raise
        except OSError as err:
            # Some port scans (e.g. nmap in -sT mode) have been known
            # to cause do_handshake to raise EBADF and ENOTCONN, so make
            # those errors quiet as well.
            # https://groups.google.com/forum/?fromgroups#!topic/python-tornado/ApucKJat1_0
            if self._is_connreset(err) or err.args[0] in (errno.EBADF, errno.ENOTCONN):
                return self.close(exc_info=True)
            raise
        except AttributeError:
            # On Linux, if the connection was reset before the call to
            # wrap_socket, do_handshake will fail with an
            # AttributeError.
            return self.close(exc_info=True)
        else:
            self._ssl_accepting = False
            if not self._verify_cert(self.socket.getpeercert()):
                self.close()
                return
            try:
                self._run_ssl_connect_callback()  # Tornado < 6
            except AttributeError:
                self._finish_ssl_connect()


class AutoHTTPSServer(HTTPServer):
    # HTTPServer implementing protocol switching.

    def _handle_connection(self, connection, address):
        # Copy-paste of tornado.httpserver.HTTPServer._handle_connection to use
        # our custom stream class.
        #
        # Actually, connection is just a socket.
        try:
            context = ssl_options_to_context(self.ssl_options)
        except Exception as e:
            logger.error("Failed to setup SSL context: %s", e)
            logger.error("Check SSL configuration and files.")
            return connection.close()

        try:
            connection = context.wrap_socket(
                connection, server_side=True, do_handshake_on_connect=False
            )
        except ssl.SSLError as err:
            if err.args[0] == ssl.SSL_ERROR_EOF:
                return connection.close()
            else:
                raise
        except OSError as err:
            # If the connection is closed immediately after it is created
            # (as in a port scan), we can get one of several errors.
            # wrap_socket makes an internal call to getpeername,
            # which may return either EINVAL (Mac OS X) or ENOTCONN
            # (Linux).  If it returns ENOTCONN, this error is
            # silently swallowed by the ssl module, so we need to
            # catch another error later on (AttributeError in
            # SSLIOStream._do_ssl_handshake).
            # To test this behavior, try nmap with the -sT flag.
            # https://github.com/tornadoweb/tornado/pull/750
            if errno_from_exception(err) in (errno.ECONNABORTED, errno.EINVAL):  # noqa
                return connection.close()
            else:
                raise
        try:
            io_loop = self.io_loop
            kw = dict(io_loop=io_loop)
        except AttributeError:
            # We are on Tornado 5+. Just don't pass ioloop
            kw = {}
            io_loop = IOLoop.current()

        try:
            stream = EasySSLIOStream(
                connection,
                max_buffer_size=self.max_buffer_size,
                read_chunk_size=self.read_chunk_size,
                **kw,
            )
            future = self.handle_stream(stream, address)
            if future is not None:
                io_loop.add_future(future, lambda f: f.result())
        except Exception:
            app_log.error("Error in connection callback", exc_info=True)

    @gen.coroutine
    def handle_stream(self, ssl_stream, address):
        try:
            yield ssl_stream.wait_for_handshake()
        except SSLErrorHTTPRequest:
            stream = make_iostream(ssl_stream)
            conn = HTTP1Connection(stream, is_client=False)
            try:
                yield self.handle_http_connection(conn)
            except Exception:
                logger.exception("Failed to process HTTP request:")
            finally:
                stream.close()
        except ssl.SSLError as e:
            logger.error("SSLError: %s", e)
        except (OSError, StreamClosedError) as e:
            logger.debug(
                "Error during SSL handshake: %s %s",
                e.__class__.__name__,
                getattr(e, "strerror", str(e)),
            )
            logger.debug("Stream closed by client during handshake. Skipping.")
            return
        else:
            super().handle_stream(ssl_stream, address)

    @gen.coroutine
    def handle_http_connection(self, conn):
        # Read the trailing HTTP request and process it with protocol_switcher.
        # We can't rely on ioloop to trigger read because it has been already
        # triggered for SSL handshake.
        addr, port = conn.stream.socket.getsockname()
        try:
            # This is not blocking. Just read available bytes.
            payload = conn.stream.socket.recv(1024)
        except Exception:
            # Exception includes EWOULDBLOCK, when no bytes are available. In
            # this case just skip.
            payload = ""
        else:
            logger.debug("Received %r", payload[:128])
        # Simulate conn._read_message side effect. This is required by
        # HTTP1Connection.write_headers()
        conn._request_start_line = parse_request_start_line("GET / HTTP/1.1")
        conn._request_headers = HTTPHeaders()
        try:
            start_line, headers = parse_http_headers(payload)
            conn._request_start_line = start_line
            request = HTTPRequest(
                connection=conn, headers=headers, start_line=start_line
            )
            request.config = self.request_callback.config
            response = protocol_switcher(request)
        except Exception as e:
            logger.error("Failed to switch to HTTPS: %s", e)
            response = HTTPResponse(
                request=object(),
                code=500,
                headers=HTTPHeaders({"Content-Length": "0"}),
                effective_url="https://useless_effective_url",
            )
        yield conn.write_headers(
            start_line=ResponseStartLine("HTTP/1.1", response.code, response.reason),
            headers=response.headers,
        )


def make_iostream(ssl_stream):
    # Detach FD from SSL socket and recreate a regular socket.
    fd = ssl_stream.socket.detach()
    ssl_stream.io_loop.remove_handler(fd)
    sock = socket.socket(
        family=ssl_stream.socket.family,
        type=ssl_stream.socket.type,
        proto=ssl_stream.socket.proto,
        fileno=fd,
    )
    return IOStream(sock)



import string
from secrets import choice

import sqlalchemy
from sqlalchemy import Column, Table, orm, schema, text, types
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Query, relationship

from ..toolkit.utils import utcnow
from . import QUERIES

Model = declarative_base()

# For bridging ORM with raw SQL code, see
# https://docs.sqlalchemy.org/en/14/orm/queryguide.html#getting-orm-results-from-textual-and-core-statements


class ApiKey(Model):
    __tablename__ = "apikeys"
    __table_args__ = {"schema": "application"}

    id = Column(types.BigInteger, primary_key=True)
    secret = Column(types.UnicodeText)
    comment = Column(types.UnicodeText)
    cdate = Column(types.TIMESTAMP(timezone=True))
    edate = Column(types.TIMESTAMP(timezone=True))

    _SECRET_LETTERS = string.ascii_letters + string.digits + "+/-_"

    @classmethod
    def generate_secret(cls, length=40):
        return "".join(choice(cls._SECRET_LETTERS) for _ in range(length))

    @classmethod
    def insert(cls, secret, comment):
        return Query(cls).from_statement(
            text(QUERIES["apikeys-insert"]).bindparams(secret=secret, comment=comment)
        )

    @classmethod
    def select_active(cls):
        return Query(cls).from_statement(text(QUERIES["apikeys-select-active"]))

    @classmethod
    def delete(cls, id):
        return Query(cls).from_statement(
            text(QUERIES["apikeys-delete"]).bindparams(id=id)
        )

    @classmethod
    def purge(cls):
        return Query(cls).from_statement(text(QUERIES["apikeys-purge"]))

    @classmethod
    def select_secret(cls, secret):
        return Query(cls).from_statement(
            text(QUERIES["apikeys-select-secret"]).bindparams(secret=secret)
        )

    @property
    def expired(self):
        return self.edate < utcnow()


class Plugin(Model):
    __tablename__ = "plugins"
    __table_args__ = (
        schema.PrimaryKeyConstraint("agent_address", "agent_port", "plugin_name"),
        schema.ForeignKeyConstraint(
            ["agent_address", "agent_port"],
            ["application.instances.agent_address", "application.instances.agent_port"],
        ),
        {"schema": "application"},
    )

    agent_address = Column(types.UnicodeText)
    agent_port = Column(types.Integer)
    plugin_name = Column(types.UnicodeText)

    instance = relationship("Instance", back_populates="plugins")

    @classmethod
    def insert(cls, instance, name):
        return Query(cls).from_statement(
            text(QUERIES["instance-enable-plugin"]).bindparams(
                agent_address=instance.agent_address,
                agent_port=instance.agent_port,
                name=name,
            )
        )

    @classmethod
    def delete(cls, instance, name):
        return text(QUERIES["instance-disable-plugin"]).bindparams(
            address=instance.agent_address, port=instance.agent_port, name=name
        )


memberships = Table(
    "memberships",
    Model.metadata,
    Column("id", primary_key=True),
    Column("role_name", schema.ForeignKey("application.roles.role_name")),
    Column("group_id", schema.ForeignKey("application.groups.id")),
    schema="application",
)


class Role(Model):
    __tablename__ = "roles"
    __table_args__ = {"schema": "application"}

    role_name = Column(types.UnicodeText, primary_key=True)
    role_password = Column(types.UnicodeText)
    role_email = Column(types.UnicodeText)
    role_phone = Column(types.UnicodeText)
    is_active = Column(types.Boolean)
    is_admin = Column(types.Boolean)

    groups = relationship(
        "Group", secondary=memberships, back_populates="members", lazy="raise"
    )

    @classmethod
    def count(cls):
        return text(QUERIES["roles-count"]).columns(count=types.Integer)

    @classmethod
    def all(cls):
        return (
            Query(cls)
            .from_statement(
                text(QUERIES["roles-all"]).columns(
                    Role.role_name,
                    Role.role_email,
                    Role.role_phone,
                    Role.is_active,
                    Role.is_admin,
                    Group.id,
                    Group.name,
                    Environment.id,
                    Environment.name,
                )
            )
            .options(orm.contains_eager(cls.groups).contains_eager(Group.environment))
        )

    @classmethod
    def delete(cls, name):
        return text(
            """DELETE FROM application.roles WHERE role_name = :name;"""
        ).bindparams(name=name)

    @classmethod
    def get(cls, name):
        return (
            Query(cls)
            .from_statement(
                text(QUERIES["roles-get"])
                .bindparams(name=name)
                .columns(
                    Role.role_name,
                    Role.role_email,
                    Role.role_phone,
                    Role.is_active,
                    Role.is_admin,
                    Group.id,
                    Group.name,
                    Environment.id,
                    Environment.name,
                )
            )
            .options(orm.contains_eager(cls.groups).contains_eager(Group.environment))
        )

    def asdict(self):
        return dict(
            name=self.role_name,
            email=self.role_email,
            phone=self.role_phone,
            active=self.is_active,
            admin=self.is_admin,
            groups=[g.name for g in self.groups],
            environments=[g.environment.name for g in self.groups],
        )

    def select_environments(self):
        return Query(Environment).from_statement(
            text(QUERIES["roles-select-environments"])
            .bindparams(role_name=self.role_name)
            .columns(Environment.id, Environment.name)
        )

    def select_instances(self):
        return Query(Instance).from_statement(
            text(QUERIES["roles-select-instances"])
            .bindparams(role_name=self.role_name)
            .columns(Instance.__mapper__.c.values())
        )


class StubRole:
    # Fake object for roles not in database.
    def __init__(self, role_name):
        self.role_name = role_name


class Group(Model):
    __tablename__ = "groups"
    __table_args__ = {"schema": "application"}

    id = Column(types.BigInteger, primary_key=True)
    name = Column(types.UnicodeText)
    description = Column(types.UnicodeText)

    members = relationship(
        "Role", secondary=memberships, back_populates="groups", lazy="raise"
    )
    environment = relationship(
        "Environment", back_populates="dba_group", uselist=False, lazy="raise"
    )

    @classmethod
    def get(cls, name):
        return Query(cls).from_statement(
            text(QUERIES["group-get"]).bindparams(name=name)
        )

    @classmethod
    def delete(cls, name):
        return text(
            """DELETE FROM application.groups WHERE name = :name;"""
        ).bindparams(name=name)

    @classmethod
    def select_membership(self, name, username):
        return (
            text(QUERIES["group-select-membership"])
            .bindparams(group=name, role=username)
            .columns(
                username=types.UnicodeText,
                groupname=types.UnicodeText,
                description=types.UnicodeText,
            )
        )

    def insert_member(self, username):
        return text(QUERIES["group-insert-member"]).bindparams(
            group=self.name, role=username
        )

    @classmethod
    def delete_member(self, name, username):
        return text(QUERIES["group-delete-membership"]).bindparams(
            group=name, role=username
        )


class Environment(Model):
    __tablename__ = "environments"
    __table_args__ = {"schema": "application"}

    id = Column(types.BigInteger, primary_key=True)
    name = Column(types.UnicodeText)
    description = Column(types.UnicodeText)
    color = Column(types.UnicodeText)
    dba_group_id = Column(types.BigInteger, schema.ForeignKey("application.groups.id"))

    instances = relationship(
        "Instance",
        back_populates="environment",
        cascade="save-update, merge, delete, delete-orphan",
        lazy="raise",
    )
    dba_group = relationship(
        "Group", back_populates="environment", uselist=False, lazy="raise"
    )

    def __str__(self):
        return self.name

    @classmethod
    def get(cls, name):
        return (
            Query(cls)
            .from_statement(
                text(QUERIES["environments-get"])
                .bindparams(name=name)
                # Declare columns for disambiguation of id, name and description.
                .columns(*cls.__table__.c.values(), *Group.__table__.c.values())
            )
            .options(orm.contains_eager(cls.dba_group))
        )

    @classmethod
    def all(cls):
        return (
            Query(cls)
            .from_statement(text(QUERIES["environments-all"]))
            .options(orm.contains_eager(cls.dba_group))
        )

    @classmethod
    def delete(cls, name):
        return text(
            """DELETE FROM application.environments WHERE name = :name;"""
        ).bindparams(name=name)

    @classmethod
    def select_memberships(self, name):
        return (
            text(QUERIES["environment-memberships"])
            .bindparams(name=name)
            .columns(
                username=types.UnicodeText,
                groupname=types.UnicodeText,
                description=types.UnicodeText,
            )
        )

    def asdict(self):
        return dict(
            name=self.name,
            description=self.description,
            color=self.color,
            dba_group=self.dba_group.name,
        )


class Instance(Model):
    __tablename__ = "instances"
    __table_args__ = (
        schema.PrimaryKeyConstraint("agent_address", "agent_port"),
        {"schema": "application"},
    )

    agent_address = Column(types.UnicodeText)
    agent_port = Column(types.Integer)
    hostname = Column(types.UnicodeText)
    pg_port = Column(types.Integer)
    notify = Column(types.Boolean, server_default=schema.FetchedValue())
    comment = Column(types.UnicodeText)
    discover = Column(postgresql.JSONB)
    discover_date = Column(types.TIMESTAMP, server_default=schema.FetchedValue())
    discover_etag = Column(types.UnicodeText)
    environment_id = Column(
        types.Integer, schema.ForeignKey("application.environments.id")
    )

    environment = relationship("Environment", back_populates="instances", lazy="raise")
    plugins = relationship(Plugin, back_populates="instance", lazy="joined")

    def __str__(self):
        return f"{self.hostname}:{self.pg_port}"

    @classmethod
    def insert(
        cls,
        agent_address,
        agent_port,
        discover,
        discover_etag,
        environment,
        notify=False,
        comment=None,
    ):
        return (
            Query(cls)
            .from_statement(
                text(QUERIES["instances-insert"]).bindparams(
                    sqlalchemy.bindparam(
                        "discover", value=discover, type_=postgresql.JSONB
                    ),
                    agent_address=str(agent_address),
                    agent_port=int(agent_port),
                    discover_etag=discover_etag,
                    discover_date=utcnow(),
                    pg_port=int(discover["postgres"]["port"]),
                    hostname=discover["system"]["fqdn"],
                    environment=environment,
                    notify=bool(notify),
                    comment=comment or "",
                )
            )
            .options(orm.contains_eager(cls.plugins))
        )

    @classmethod
    def delete(cls, agent_address, agent_port):
        return text(
            """DELETE FROM application.instances WHERE agent_address = :address AND agent_port = :port;"""
        ).bindparams(address=str(agent_address), port=int(agent_port))

    @classmethod
    def get(cls, agent_address, agent_port):
        return (
            Query(cls)
            .from_statement(
                text(QUERIES["instance-get"])
                .bindparams(address=str(agent_address), port=int(agent_port))
                .columns(
                    *cls.__mapper__.c.values(),
                    *Plugin.__mapper__.c.values(),
                    *Environment.__mapper__.c.values(),
                    *Group.__mapper__.c.values(),
                )
            )
            .options(
                orm.contains_eager(cls.plugins),
                # chain contains_eager to reflect chained joins.
                orm.contains_eager(cls.environment).contains_eager(
                    Environment.dba_group
                ),
            )
        )

    @classmethod
    def select_for_home(cls, role_name):
        return (
            text(QUERIES["instances-select-for-home"])
            .bindparams(role_name=role_name)
            .columns(
                agent_address=types.UnicodeText,
                agent_port=types.Integer,
                hostname=types.UnicodeText,
                pg_port=types.Integer,
                pg_data=types.UnicodeText,
                pg_version=types.UnicodeText,
                pg_version_summary=types.UnicodeText,
                groups=types.ARRAY(types.UnicodeText),
                plugins=types.ARRAY(types.UnicodeText),
                available=types.Boolean,
                checks=postgresql.JSONB,
            )
        )

    @classmethod
    def count(cls):
        return text(QUERIES["instances-count"]).columns(count=types.Integer)

    @classmethod
    def all(cls):
        return (
            Query(cls)
            .from_statement(text(QUERIES["instances-all"]))
            .options(orm.contains_eager(cls.environment))
        )

    # Compatibility from new JSONb discover to old column discover.
    @property
    def cpu(self):
        d = self.discover or {}
        return d.get("system", {}).get("cpu_count")

    @property
    def memory_size(self):
        d = self.discover or {}
        return d.get("system", {}).get("memory")

    @property
    def pg_data(self):
        d = self.discover or {}
        return d.get("postgres", {}).get("data_directory")

    @property
    def pg_version(self):
        d = self.discover or {}
        return d.get("postgres", {}).get("version")

    @property
    def pg_version_summary(self):
        d = self.discover or {}
        return d.get("postgres", {}).get("version_summary")

    def dashboard_url(self, app):
        scheme = "https" if app.config.temboard.ssl_key_file else "http"
        host = app.config.temboard.address
        port = app.config.temboard.port
        path = f"/server/{self.agent_address}/{self.agent_port}/dashboard"
        return f"{scheme}://{host}:{port}{path}"

    def asdict(self):
        return dict(
            hostname=self.hostname,
            pg_port=self.pg_port,
            agent_address=self.agent_address,
            agent_port=self.agent_port,
            environment=self.environment.name,
            plugins=[plugin.plugin_name for plugin in self.plugins],
            comment=self.comment,
            notify=self.notify,
            discover=self.discover,
            discover_etag=self.discover_etag,
            discover_date=self.discover_date,
        )

    def select_email_and_phone_for_notify(self):
        return (
            text(QUERIES["instance-select-email-and-phone-for-notify"])
            .bindparams(agent_address=self.agent_address, agent_port=self.agent_port)
            .columns(
                emails=types.ARRAY(types.UnicodeText),
                phones=types.ARRAY(types.UnicodeText),
            )
        )

    def has_dba(self, role_name):
        # DEPRECATED: Use ACL once implemented.
        return (
            text(QUERIES["instance-has-dba"])
            .bindparams(
                agent_address=self.agent_address,
                agent_port=self.agent_port,
                role_name=role_name,
            )
            .columns(has_dba=types.Boolean)
        )

    def enable_plugin(self, plugin):
        return Plugin.insert(self, plugin)

    def disable_plugin(self, plugin):
        return Plugin.delete(self, plugin)



import json
import logging
import os.path
import sys

from ..agentclient import TemboardAgentClient
from ..model import Session, orm
from ..toolkit import validators as v
from ..toolkit.app import SubCommand
from ..toolkit.errors import UserError
from ..toolkit.utils import JSONEncoder
from .app import app

logger = logging.getLogger(__name__)


@app.command
class RegisterInstance(SubCommand):
    """Out-of-bound registration of a PostgreSQL instance."""

    name = "register-instance"

    def define_arguments(self, parser):
        parser.add_argument(
            "--comment",
            default="Registered from CLI.",
            help="Instance comment. Default: %(default)s",
        )

        parser.add_argument(
            "--notify",
            action="store_true",
            default=False,
            help="Whether to contact users on monitoring alert.",
        )

        parser.add_argument(
            "--if-not-exists",
            action="store_true",
            dest="skip_existing",
            default=False,
            help="Don't fail if agent is already registered.",
        )

        parser.add_argument(
            "-e",
            "--environment",
            dest="environment",
            default="",
            help="Instance environment name.",
        )

        parser.add_argument(
            "-p",
            "--plugins",
            dest="plugins",
            default="",
            help=(
                "Activate this temBoard UI plugins, comma separated. "
                "Defaults to agent's plugins."
            ),
        )

        parser.add_argument(
            "agent_address", metavar="ADDR", help="temBoard agent listenning address"
        )

        parser.add_argument(
            "agent_port",
            metavar="PORT",
            nargs="?",
            default=2345,
            type=int,
            help="temBoard agent listenning port",
        )

        super().define_arguments(parser)

    def main(self, args):
        agent = f"{args.agent_address}:{args.agent_port}"

        if not args.environment:
            raise UserError("Missing environment. Use --environment to define.")

        plugins = v.commalist(args.plugins)
        for plugin in plugins:
            if plugin not in self.app.config.temboard.plugins:
                raise UserError("Plugin %s is not loaded." % plugin)

        session = Session()
        logger.debug("Check for existing instance.")
        instance = (
            orm.Instance.get(
                agent_address=args.agent_address, agent_port=args.agent_port
            )
            .with_session(session)
            .one_or_none()
        )

        if instance:
            if not args.skip_existing:
                raise UserError(
                    "Instance %s:%s already registered."
                    % (instance.hostname, instance.pg_port)
                )
            else:
                logger.warning(
                    "Instance %s:%s already registered. Skipping.",
                    instance.hostname,
                    instance.pg_port,
                )
                self.output_instance(instance)
                logger.info("Browse instance at %s.", instance.dashboard_url(self.app))
                return 0

        logger.info("Discovering temBoard agent at %s.", agent)
        client = TemboardAgentClient.factory(
            self.app.config, args.agent_address, args.agent_port
        )
        try:
            response = client.get("/discover")
            response.raise_for_status()
        except OSError as e:
            logger.error("Failed to discover agent at %s: %s", agent, e)
            raise UserError(
                "Can't connect to agent. "
                "Please check address and port or that agent is running."
            )
        except client.Error as e:
            raise UserError("temBoard Agent error: %s" % e)
        else:
            discover = response.json()
            discover_etag = response.headers["ETag"]

        if "signature_status" not in discover:
            logger.error("Legacy agent does not support out of band registration.")
            logger.error("Please upgrade agent or use web UI.")
            raise UserError("Legacy agent registration refused")

        if "valid" != discover["signature_status"]:
            logger.error(
                "Agent %s is configured for another UI or " "signing key is outdated.",
                agent,
            )
            logger.error("Update signing key with temboard-agent fetch-key.")
            raise UserError("Signature missmatch")

        logger.info(
            "Registering %s serving %s at %s:%s.",
            discover["postgres"]["version_summary"],
            discover["postgres"]["data_directory"],
            discover["system"]["hostname"],
            discover["postgres"]["port"],
        )
        if plugins:
            for plugin in plugins:
                if plugin not in discover["temboard"]["plugins"]:
                    raise UserError("Plugin %s disabled on agent side." % plugin)
        else:
            plugins = guess_default_plugins(
                ui_plugins=self.app.config.temboard.plugins,
                agent_plugins=discover["temboard"]["plugins"],
            )
        logger.debug("Enabling plugins %s.", ", ".join(plugins))

        environment = session.execute(orm.Environment.get(args.environment)).fetchone()
        instance = (
            orm.Instance.insert(
                agent_address=args.agent_address,
                agent_port=args.agent_port,
                discover=discover,
                discover_etag=discover_etag,
                notify=args.notify,
                comment=args.comment,
                environment=environment,
            )
            .with_session(session)
            .one()
        )

        for plugin in plugins:
            session.execute(instance.enable_plugin(plugin))

        session.commit()
        logger.info("Instance successfully registered.")

        logger.debug("Fast-collect monitoring metrics.")
        if os.path.exists(self.app.scheduler.scheduler.address):
            if "monitoring" in plugins:
                from ..plugins.monitoring import collector

                logger.info("Schedule monitoring collect for agent now.")
                collector.defer(
                    self.app, address=instance.agent_address, port=instance.agent_port
                )

            if "statements" in plugins:
                from ..plugins.statements import statements_pull1

                logger.info("Schedule statements collect for agent now.")
                statements_pull1.defer(
                    self.app, host=instance.agent_address, port=instance.agent_port
                )
        else:
            logger.info(
                "Scheduler is not running." "Start temBoard server to collect metrics."
            )

        self.output_instance(instance)

        logger.info("Browse instance at %s.", instance.dashboard_url(self.app))

    def output_instance(self, instance):
        json.dump(instance.asdict(), sys.stdout, indent="  ", cls=JSONEncoder)
        sys.stdout.write(os.linesep)


def guess_default_plugins(ui_plugins, agent_plugins):
    # Select all plugins enabled in both UI and agent.
    return set(ui_plugins) & set(agent_plugins)



import json
import logging
from uuid import uuid4

from .toolkit.http import TemboardClient, format_date
from .toolkit.signing import canonicalize_request, sign_v1

logger = logging.getLogger(__name__)


class TemboardAgentClient(TemboardClient):
    @classmethod
    def factory(cls, config, host, port, username="temboard"):
        return cls(
            config.signing_key,
            host,
            port,
            ca_cert_file=config.temboard.ssl_ca_cert_file,
            username=username,
        )

    def __init__(self, signing_key, host, port, ca_cert_file=None, username="temboard"):
        super().__init__(host, port, ca_cert_file)
        self.signing_key = signing_key
        self.username = username

    def request(self, method, path, headers=None, body=None):
        hostport = f"{self.host}:{self.port}"
        headers = headers or {}

        headers.setdefault("Host", hostport)
        headers.setdefault("X-TemBoard-Date", format_date())
        headers.setdefault("X-TemBoard-Request-Id", str(uuid4()))
        headers.setdefault("X-TemBoard-User", self.username)

        # Preprocess body to sign it.
        if isinstance(body, (dict, list)):
            body = json.dumps(body)
            headers["Content-Type"] = "application/json"

        if hasattr(body, "isnumeric"):  # Is it a unicode ?
            body = body.encode("utf-8")

        if "POST" == method:
            headers["Content-Length"] = str(len(body or b""))

        canonical_request = canonicalize_request(method, path, headers, body)
        if self.log_headers:
            logger.debug("Canonical request:\n%s", canonical_request.decode("utf-8"))
        signature = sign_v1(self.signing_key, canonical_request)
        headers["X-TemBoard-Signature"] = "v1:%s" % signature

        return super().request(method, path, headers, body)



from flask import current_app, jsonify
from flask import render_template as flask_render_template

from ...web.flask import instance_proxy, instance_routes


class ActivityPlugin:
    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        instance_proxy.generic_proxy("/activity/kill", method="POST")


@instance_routes.route("/activity")
def activity():
    current_app.instance.check_active_plugin("activity")
    current_app.instance.fetch_status()
    return flask_render_template("activity.html", plugin="activity")


@instance_proxy.route("/activity")
def activity_proxy():
    # DEPRECATED: move to new agent endpoints for activity.
    return jsonify(
        dict(
            blocking=current_app.instance.request("/activity/blocking").json(),
            running=current_app.instance.request("/activity").json(),
            waiting=current_app.instance.request("/activity/waiting").json(),
        )
    )



import logging

try:
    from inspect import signature
except ImportError:
    from inspect import getargspec
from textwrap import dedent

from ..model import check_schema
from ..toolkit.app import SubCommand
from ..toolkit.errors import UserError
from ..toolkit.taskmanager import FlushTasksMixin, RunTaskMixin
from .app import app

logger = logging.getLogger(__name__)


@app.command
class Tasks(SubCommand):
    """Manage background tasks."""

    def main(self, args):
        raise UserError("Missing sub-command. See --help for details.")


@Tasks.command
class Flush(FlushTasksMixin, SubCommand):
    """Flush all tasks."""


@Tasks.command
class Run(RunTaskMixin, SubCommand):
    """Run a task foreground."""

    def main(self, args):
        workers = self.iter_workers()

        if "?" == args.worker_name:
            self.print_workers(workers)
        else:
            worker, worker_args = self.compute_worker_args(workers, args)
            check_schema()
            worker(*worker_args)
        return 0


@Tasks.command
class Schedule(RunTaskMixin, SubCommand):
    """Schedule a background task."""

    def define_arguments(self, parser):
        parser.description = dedent("""\
        Schedule a task background.

        You need a running temBoard UI to execute this command.
        """)

        parser.add_argument(
            "worker_name",
            metavar="WORKER",
            help=(
                "Global name of the worker function name to execute."
                " Use ? to list available workers."
            ),
        )

        parser.add_argument(
            "worker_args",
            nargs="*",
            metavar="ARG",
            default=[],
            help="Worker arguments as Python literals.",
        )

    def main(self, args):
        workers = self.iter_workers()

        if "?" == args.worker_name:
            self.print_workers(workers)
        else:
            worker, worker_args = self.compute_worker_args(workers, args)
            check_schema()
            if not self.app.scheduler.can_schedule():
                raise UserError("temBoard is not running.")

            out = worker.defer(self.app, **build_kwargs_from_args(worker, worker_args))
            logger.info(
                "Worker %s scheduled with task ID %s.",
                args.worker_name,
                out.content["id"],
            )

        return 0


def build_kwargs_from_args(callable_, args):
    try:
        sig = signature(callable_).parameters
    except NameError:
        sig = getargspec(callable_).args
    names = list(sig.keys())
    names.pop(0)  # Drop app arg
    return dict(zip(names, args))



import logging

from temboardui.web.tornado import admin_required, app, render_template

from ...version import inspect_versions

logger = logging.getLogger(__name__)


@app.route(r"/settings/metadata")
@admin_required
def metadata(request):
    versions_info = inspect_versions()
    infos = {
        "Version": "%(temboard)s (%(temboardbin)s)" % versions_info,
        "Uptime": app.start_time,
        "OS": "%(distname)s %(distversion)s" % versions_info,
        "Python": "%(python)s (%(pythonbin)s)" % versions_info,
        "cryptography": "%(cryptography)s" % versions_info,
        "Tornado": "%(tornado)s" % versions_info,
        "libpq": "%(libpq)s" % versions_info,
        "psycopg2": "%(psycopg2)s" % versions_info,
        "SQLAlchemy": "%(sqlalchemy)s" % versions_info,
    }
    temboard_version = "%(temboard)s" % versions_info

    return render_template(
        "settings/metadata.html",
        role=request.current_user,
        infos=infos,
        temboard_version=temboard_version,
    )



import codecs
import logging
from io import StringIO

import flask
import sqlalchemy
from flask import current_app, g

from ... import agentclient
from ...model import QUERIES, orm
from ...toolkit import validators
from ...toolkit.utils import utcnow
from ..flask import admin_required, transaction, validating

logger = logging.getLogger(__name__)


@current_app.route("/json/environments")
@admin_required
def get_environments():
    return flask.jsonify(
        [e.asdict() for e in orm.Environment.all().with_session(g.db_session)]
    )


@current_app.route("/json/environments", methods=["POST"])
@admin_required
@transaction
def post_environments():
    return put_environment(environment=orm.Environment(dba_group=orm.Group()))


@current_app.route("/json/environments/<name>")
@admin_required
def get_environment(name):
    environment = orm.Environment.get(name).with_session(g.db_session).first()
    if environment is None:
        flask.abort(404, "No such environment.")
    return flask.jsonify(environment.asdict())


@current_app.route("/json/environments/<name>/members")
@admin_required
def get_environment_members(name):
    return flask.jsonify(
        [
            {k: getattr(r, k) for k in r.keys()}
            for r in g.db_session.execute(orm.Environment.select_memberships(name))
        ]
    )


@current_app.route("/json/environments/<name>", methods=["PUT"])
@admin_required
@transaction
def put_environment(name=None, environment=None):
    if environment is None:
        environment = orm.Environment.get(name).with_session(g.db_session).first()
    if environment is None:
        flask.abort(404, "No such environment.")

    with validating():
        environment.name = validators.slug(flask.request.json["name"])
    environment.description = flask.request.json["description"]
    environment.dba_group.name = f"{environment.name}/dba"
    # Used as profile name in /settings/environment/<>/members
    environment.dba_group.description = "DBA"
    g.db_session.add(environment)  # When called from post_environment
    g.db_session.flush()

    return flask.jsonify(environment.asdict())


@current_app.route("/json/environments/<name>", methods=["DELETE"])
@admin_required
@transaction
def delete_environment(name):
    # Delete DBA group, cascding to environment.
    result = g.db_session.execute(orm.Group.delete(f"{name}/dba"))
    if result.rowcount == 0:
        flask.abort(404, "No such environment.")
    return flask.jsonify()


@current_app.route("/json/instances", methods=["POST"])
@transaction
def post_instance():
    j = flask.request.json
    environment = (
        orm.Environment.get(j["environment"]).with_session(g.db_session).first()
    )
    if environment is None:
        flask.abort(400, "Unknown environment")
    try:
        instance = (
            orm.Instance.insert(
                agent_address=j["agent_address"],
                agent_port=j["agent_port"],
                discover=j["discover"],
                discover_etag=j["discover_etag"],
                environment=j["environment"],
                # put_instance will handle other attributes.
            )
            .with_session(g.db_session)
            .one()
        )
    except sqlalchemy.exc.IntegrityError as e:
        logger.warning("Failed to insert instance: %s", e)
        flask.abort(400, "Instance already registered.")

    # insert does not populate environment nor plugins.
    instance.environment = environment

    g.db_session.flush()
    return put_instance(instance=instance)


@current_app.route("/json/instances/<address>/<port>")
@admin_required
def get_instance(address, port):
    try:
        instance = orm.Instance.get(address, port).with_session(g.db_session).one()
    except sqlalchemy.orm.exc.NoResultFound:
        flask.abort(404, "Instance not found.")
    return flask.jsonify(instance.asdict())


@current_app.route("/json/instances/<address>/<port>", methods=["PUT"])
@admin_required
@transaction
def put_instance(address=None, port=None, instance=None):
    if not instance:
        try:
            instance = orm.Instance.get(address, port).with_session(g.db_session).one()
        except sqlalchemy.orm.exc.NoResultFound:
            flask.abort(404, "Instance not found.")

    j = flask.request.json

    try:
        instance.discover = j["discover"]
        instance.discover_etag = j["discover_etag"]
        instance.discover_date = utcnow()
        instance.hostname = j["discover"]["system"]["fqdn"]
        instance.pg_port = j["discover"]["postgres"]["port"]
    except KeyError:
        logger.debug("Missing discover. Keeping old informations.")
    instance.notify = j["notify"]
    instance.comment = j["comment"] or ""
    if j["environment"] != instance.environment.name:
        instance.environment = (
            orm.Environment.get(j["environment"]).with_session(g.db_session).first()
        )
    if instance.environment is None:
        flask.abort(400, "Unknown environment.")
    g.db_session.add(instance)
    g.db_session.flush()

    current_plugins = {p.plugin_name for p in instance.plugins}
    new_plugins = set(j["plugins"])
    for plugin in current_plugins:
        if plugin not in new_plugins:
            g.db_session.execute(instance.disable_plugin(plugin))
    for plugin in new_plugins - current_plugins:
        if plugin not in current_app.temboard.plugins:
            flask.abort(400, f"Unknown plugin {plugin}.")
        g.db_session.execute(instance.enable_plugin(plugin))

    return flask.jsonify(instance.asdict())


@current_app.route("/json/instances/<address>/<port>", methods=["DELETE"])
@admin_required
@transaction
def delete_instance(address, port):
    out = g.db_session.execute(orm.Instance.delete(address, port))
    if out.rowcount == 0:
        flask.abort(404, "Instance not found.")
    return flask.jsonify()


# Special proxy for unregistered instance.
@current_app.route("/json/instances/<address>/<port>/discover")
@admin_required
def discover(address, port):
    client = agentclient.TemboardAgentClient.factory(
        current_app.temboard.config, address, port, username=g.current_user.role_name
    )
    try:
        response = client.get("/discover")
        response.raise_for_status()
    except OSError as e:
        logger.warning("Failed to discover agent at %s:%s: %s", address, port, e)
        flask.abort(
            401,
            "Can't connect to agent. "
            "Please check address and port or that agent is running.",
        )
    return flask.jsonify(response.json())


@current_app.route("/instances.csv")
@admin_required
def get_instances_csv():
    search = flask.request.args.get("filter")
    pattern = "%%%s%%" % search if search else "%"
    bind = g.db_session.get_bind()
    conn = bind.raw_connection().connection
    sql = QUERIES["instances-copy-as-csv"]
    with conn.cursor() as cur, StringIO() as fo:
        sql = cur.mogrify(sql, (pattern,))
        cur.copy_expert(sql, fo)
        csv = fo.getvalue()

    filename = "postgresql-instances-inventory.csv"
    return flask.make_response(
        # BOM declares UTF8 for Excel.
        codecs.BOM_UTF8 + csv.encode("utf-8"),
        200,
        {
            "Content-Type": "text/csv",
            "Content-Disposition": "attachment;filename=" + filename,
        },
    )



from flask import current_app as app
from flask import g, render_template, request

from ...model import orm
from ...version import inspect_versions


@app.route("/home")
def home():
    role = g.current_user
    environments = [
        e.name for e in role.select_environments().with_session(g.db_session).all()
    ]
    return render_template("home.html", environments=environments)


@app.route("/about")
def about():
    versions_info = inspect_versions()
    instances = g.db_session.scalar(orm.Instance.count())
    roles = g.db_session.scalar(orm.Role.count())
    infos = {
        "Browser": request.headers.get("User-Agent", "Unknown"),
        "Version": "%(temboard)s (%(temboardbin)s)" % versions_info,
        "Uptime": app.start_time,
        "OS": "%(distname)s %(distversion)s" % versions_info,
        "Python": "%(python)s (%(pythonbin)s)" % versions_info,
        "cryptography": versions_info["cryptography"],
        "Tornado": versions_info["tornado"],
        "libpq": versions_info["libpq"],
        "psycopg2": versions_info["psycopg2"],
        "SQLAlchemy": versions_info["sqlalchemy"],
        "Instances": instances,
        "Users": roles,
        "SMTP": app.temboard.config.notifications.smtp_host is not None,
        "Twilio": app.temboard.config.notifications.twilio_account_sid is not None,
    }
    temboard_version = versions_info["temboard"]

    return render_template("about.html", infos=infos, temboard_version=temboard_version)



from flask import current_app, render_template

from ...web.flask import instance_routes


@instance_routes.route("/pgconf/configuration", methods=["GET"])
def pgconf_configuration():
    current_app.instance.check_active_plugin("pgconf")
    current_app.instance.fetch_status()

    return render_template("configuration.html", plugin="pgconf")



import base64
import json
import logging
import re
from binascii import hexlify
from email.mime.text import MIMEText
from hashlib import sha512
from smtplib import SMTP, SMTP_SSL
from urllib.error import HTTPError
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from sqlalchemy.orm.exc import NoResultFound

from temboardui.errors import TemboardUIError
from temboardui.model.orm import Instance, Role

logger = logging.getLogger(__name__)

"""
Instances
"""


def get_instance(session, agent_address, agent_port):
    return Instance.get(agent_address, agent_port).with_session(session).first()


"""
Roles
"""


def get_role_by_auth(session, role_name, role_password):
    try:
        role = (
            session.query(Role)
            .filter(Role.role_name == str(role_name), Role.is_active.is_(True))
            .one()
        )
        if role.role_password != str(role_password):
            raise TemboardUIError(
                400, "Wrong user/password: %s/%s" % (role_name, role_password)
            )
        return role
    except NoResultFound:
        raise TemboardUIError(
            400, "Wrong user/password: %s/%s" % (role_name, role_password)
        )


def hash_password(username, password):
    """
    Hash a password with the following formula:
        sha512(password + sha512(username))
    """
    bytes_password = hexlify(password.encode("utf-8"))
    bytes_sha_username = sha512(username.encode("utf-8")).digest()
    hpasswd = sha512(bytes_password + bytes_sha_username).digest()
    return base64.b64encode(hpasswd)


def gen_cookie(username, hash_password):
    """
    Build secure cookie content as a string containing:
        - content length (excluding length itself)
        - role_name
        - 16 first chars of the hash password

    Part of hash password is there for 2 main reasons:

        1/ If the cookie secret key is stolen, the attacker will be able to
        build a valid secure cookie, so in this situation having a
        non-predictibable variable - part of hash password -  wich can be
        checked on server side will render the attack effective less, unless
        if the attacker knows role's password.

        2/ We don't want to see the full hash password stored in a cookie.
    """
    content = f"{username}:{hash_password[:15]}"
    return f"{len(content) + 1}:{content}"


def get_role_by_cookie(session, content):
    p_cookie_content = r"^([0-9]+):([a-z0-9_\-.]{3,16}):([A-Za-z0-9+/]{15})$"
    r_cookie_content = re.compile(p_cookie_content)
    match = r_cookie_content.match(content)
    if match:
        c_length = match.group(1)
        c_role_name = match.group(2)
        c_fp_hash_password = match.group(3)
        if int(c_length) != (len(c_role_name) + len(c_fp_hash_password) + 2):
            raise Exception("Cookie's content is corrupted.")
        try:
            role = (
                session.query(Role)
                .filter(Role.role_name == str(c_role_name), Role.is_active.is_(True))
                .one()
            )
        except (NoResultFound, Exception):
            raise Exception("Role '%s' not found or not active." % (c_role_name))
        if role.role_password[:15] != str(c_fp_hash_password):
            raise Exception("Password sign not correct.")
        return role
    else:
        raise Exception("Cookie content is not valid.")


def send_mail(
    host,
    port,
    subject,
    content,
    emails,
    tls=False,
    login=None,
    password=None,
    from_addr=None,
):
    msg = MIMEText(content, "plain", "utf-8")
    msg["Subject"] = subject

    try:
        if tls:
            smtp = SMTP_SSL(host, port)
        else:
            smtp = SMTP(host, port)

        if login and password:
            logger.debug("Authenticating to SMTP server as %s.", login)
            smtp.login(login, password)

        smtp.sendmail(from_addr, emails, msg.as_string())
        smtp.quit()
    except Exception as e:
        raise TemboardUIError(
            500, "Could not send mail; %s\n" "SMTP connection may be misconfigured." % e
        )


def send_sms(config, content, phones):
    sid = config.twilio_account_sid
    token = config.twilio_auth_token
    from_ = config.twilio_from
    uri = "https://api.twilio.com/2010-04-01/Accounts/%s/Messages.json" % sid
    s = base64.b64encode(f"{sid}:{token}")

    errors = []
    for recipient in phones:
        req = Request(url=uri)
        req.add_header("Authorization", "Basic %s" % s)
        data = {"From": from_, "Body": content, "To": recipient}
        req.add_data(urlencode(data))
        try:
            urlopen(req)
        except HTTPError as e:
            response = json.loads(e.read())
            logger.error("Could not send SMS; %s" % response.get("message"))
            errors.append(recipient)
        except Exception as e:
            logger.error("Could not send SMS; %s" % e)
            errors.append(recipient)

    if errors:
        raise TemboardUIError(
            500,
            "Could not send SMS to %s; \n See logs for more information"
            % ", ".join(errors),
        )
